/* msx-emul is (C) 2008-2009 by folkert@vanheusden.com
 * released under GPL
 */
static char *rcsId = "$Id: disk.c,v 1.26 2009/02/18 14:16:50 folkert Exp folkert $";

#include <stdio.h>
#include <string.h>
#include <ncurses.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "error.h"
#include "msx.h"

#define T1_BUSY		0x01
#define T1_INDEX	0x02
#define T1_TRACK0	0x04
#define T1_CRCERR	0x08
#define T1_SEEKERR	0x10
#define T1_HEADLOAD	0x20
#define T1_PROT		0x40
#define T1_NOTREADY	0x80

#define T2_BUSY		0x01
#define T2_DRQ		0x02
#define T2_NOTREADY	0x80

extern char verbose_disk;

int disk_slot = -1, disk_sub_slot = 0;
char *disk_file;
int disk_fh = -1;
int track = 0;
int drive;
int io_buffer_index = 0, io_buffer_left = 0;
char last_step_dir = 0;
unsigned char flags = 0;
unsigned char disk_reg[16];
unsigned char io_buffer[16384]; /* enough to buffer a track of 9 sectors and some overhead */
int busy_cnt = 0;

typedef enum { type_i, type_ii, type_iii, type_iv } typecommand_t;
typedef enum { st_read, st_write, st_idle } rw_state_t;
typedef enum { wm_sector, wm_track } wm_mode_t;

typecommand_t tc = type_i;
rw_state_t rws = st_idle;
wm_mode_t write_mode = wm_sector;

void open_disk(char *file, int *fh)
{
	if (*fh == -1)
	{
		off_t offset;

		*fh = open(file, O_RDWR);
		if (*fh == -1)
			error_exit("error opening disk file %s", file);

		offset = lseek(*fh, 0, SEEK_END);
		if (offset == -1)
			error_exit("seek error in file\n");

		if (offset == 0)
		{
			int loop;

			for(loop=0; loop<(80*9*2); loop++)
			{
				if (write(*fh, io_buffer, 512) != 512)
					error_exit("error writing to file");
			}

			if (lseek(*fh, 0, SEEK_SET) == -1)
				error_exit("seek error in file\n");
		}
	}
}

int file_offset(int side, int track, int sector)
{
	int offset;

	offset = (sector - 1)* 512 + (track * 9 * 512) + (80 * 9 * 512) * side;

	if (verbose_disk) fprintf(stderr, "file offset side %d track %d sector %d: %d\n", side, track, sector, offset);

	return offset;
}

int get_side(unsigned char *disk_reg)
{
	return disk_reg[0x0c] & 10 ? 1 : 0;
}

int go_to_track(int track)
{
	int side   = get_side(disk_reg);

	if (lseek(disk_fh, file_offset(side, track, 1), SEEK_SET) == -1)
	{
		if (verbose_disk) fprintf(stderr, "go_to_track: seek error in file\n");
		return -1;
	}

	return 0;
}

int go_to_sector(unsigned char disk_reg[16], int track)
{
	int sector = disk_reg[0x0a];
	int side   = get_side(disk_reg);
	int go_to_offset = file_offset(side, track, sector);
	int new_offset = lseek(disk_fh, go_to_offset, SEEK_SET);

	if (verbose_disk) fprintf(stderr, "go_to_sector: goto: %d, real: %d\n", go_to_offset, new_offset);

	if (new_offset == -1)
	{
		if (verbose_disk) fprintf(stderr, "go_to_sector: seek error in file\n");
		return -1;
	}

	return 0;
}

void debug_show_disk_pos(void)
{
	int pos = lseek(disk_fh, 0, SEEK_CUR), side, track, sector, pos2 = pos;
	side = pos / (80 * 9 * 512);
	pos -= side * 80 * 9 * 512;
	track = pos / (9 * 512);
	pos -= track * 9 * 512;
	sector = (pos  / 512) + 1;

	fprintf(stderr, "offset in file: side %d track %d sector %d (%d)\n", side, track, sector, pos2);
}

unsigned char disk_memorymapped_io_read(int address)
{
	if (disk_slot == -1) { if (verbose_disk) fprintf(stderr, "DISK mem-i/o read: no disk rom loaded?!"); }

	open_disk(disk_file, &disk_fh);

	mvwprintw(stdscr, 10, 0, "regR %x", address & 0xf);

	if (address == 0x08)	/* status register */
	{
		if (tc == type_i || tc == type_iv)
		{
			unsigned char data = flags;

			if (verbose_disk) fprintf(stderr, "read status register type i: %02x\n", flags);
			flags &= T1_NOTREADY;
			flags &= T1_BUSY;

			return data;
		}
		else if (tc == type_ii || tc == type_iii)
		{
			unsigned char data = flags;

			if (verbose_disk) fprintf(stderr, "read status register type ii: %02x\n", data);

			//			flags &= ~(T2_BUSY | T2_DRQ);

			return data;
		}

		if (verbose_disk) fprintf(stderr, "status reg: unknown state\n");
	}
	else if (address == 0x09)			/* track */
	{
		if (verbose_disk) fprintf(stderr, "read track nr %d\n", disk_reg[address]);
		return disk_reg[address];
	}
	else if (address == 0x0a)			/* sector */
	{
		if (verbose_disk) fprintf(stderr, "read sector nr %d\n", disk_reg[address]);
		return disk_reg[address];
	}
	else if (address == 0x0b)	/* data register */
	{
		unsigned char data = -1;

		if (verbose_disk && io_buffer_index == 0) fprintf(stderr, "start read data %02x at index %d\n", io_buffer[io_buffer_index], io_buffer_index);

		if (rws == st_read)
		{
			data = io_buffer[io_buffer_index++];
			io_buffer_left--;

			if (io_buffer_left == 0)
			{
				if (verbose_disk) fprintf(stderr, "finished reading data at index %d\n", io_buffer_index);

				rws = st_idle;
				io_buffer_index = 0;
				io_buffer_left = 0;
				flags &= ~(T2_DRQ | T2_BUSY | 32/*fault/type*/);
			}
			else
			{
				flags |= T2_DRQ;
			}
		}
		else
		{
			if (verbose_disk) fprintf(stderr, "read sector already finished reading (offset %d)\n", io_buffer_index);
		}

		return data;
	}
	else if (address == 0x0c)			/* side */
	{
		if (verbose_disk) fprintf(stderr, "read side nr %d\n", disk_reg[address]);
		return disk_reg[address];
	}
	else if (address == 0x0f)			/* DRQ (bit 7) */
	{
		unsigned char data = 0;

		if (flags & T2_DRQ)
		{
			data |= 128;
			flags &= ~T2_DRQ;
		}
		if (flags & T2_BUSY)
		{
			data |= 64;
			//			flags &= ~T2_BUSY;	must not clear this bit here
		}

		if (verbose_disk > 1) fprintf(stderr, "read DRQ: %d\n", data);

		return data;
	}
	else if (address == 0x0d)			/* motor */
	{
		if (verbose_disk) fprintf(stderr, "read motor state %d\n", disk_reg[address]);

		return disk_reg[address];
	}
	else
	{
		if (verbose_disk) fprintf(stderr, "unhandled disk read: %04x\n", address);
	}

	return -1;
}

void disk_memorymapped_io_write(int address, unsigned char value)
{
	if (disk_slot == -1) { if (verbose_disk) fprintf(stderr, "DISK mem-i/o write: no disk rom loaded?!"); }

	open_disk(disk_file, &disk_fh);

	disk_reg[address] = value;

	if (address == 0x08)	/* command register */
	{
		int command = value >> 4;
		char T      = (value >> 4) & 1;
		char h      = (value >> 3) & 1;
		char V      = (value >> 2) & 1;
		char r1     = (value >> 1) & 1;
		char r0     = (value     ) & 1;
		char m      = (value >> 4) & 1;
		char S      = (value >> 3) & 1;
		char E      = (value >> 2) & 1;
		char C      = (value >> 1) & 1;
		char A0     = (value     ) & 1;
		char i      = (value & 15);

		if (verbose_disk) fprintf(stderr, "write command register %02x: %02x, cmd %d\n", address, value, command);

		if (command == 0)	/* I restore */
		{
			if (verbose_disk) fprintf(stderr, "i restore\n");
			track = disk_reg[0x09] = 0;

			flags = T1_INDEX | T1_TRACK0;
			if (h) flags |= T1_HEADLOAD;

			tc = type_i;
		}
		else if (command == 1)	/* I seek */
		{
			if (verbose_disk) fprintf(stderr, "i seek to %d\n", disk_reg[0x0b]);
			track = disk_reg[0x09] = disk_reg[0x0b]; /* new track is in data register */

			flags = T1_INDEX | T1_TRACK0;
			if (h) flags |= T1_HEADLOAD;

			tc = type_i;
		}
		else if (command == 2 || command == 3)	/* I step */
		{
			if (verbose_disk) fprintf(stderr, "i step to %d + %d\n", track, last_step_dir);
			track += last_step_dir;

			if (track < 0) track = 0;
			if (track > 76) track = 76;

			flags = T1_INDEX;
			if (track == 0) flags |= T1_TRACK0;

			if (T) disk_reg[0x09] = track;

			tc = type_i;
		}
		else if (command == 4 || command == 5)	/* I step in */
		{
			track++;
			if (verbose_disk) fprintf(stderr, "i step in to %d\n", track);
			if (track > 79) track = 79;

			last_step_dir = 1;

			flags = T1_INDEX;
			if (track == 0) flags |= T1_TRACK0;

			if (T) disk_reg[0x09] = track;

			tc = type_i;
		}
		else if (command == 6 || command == 7)	/* I step out */
		{
			track--;
			if (verbose_disk) fprintf(stderr, "i step out to %d\n", track);
			if (track < 0) track = 0;

			last_step_dir = -1;

			flags = T1_INDEX;
			if (track == 0) flags |= T1_TRACK0;

			if (T) disk_reg[0x09] = track;

			tc = type_i;
		}
		else if (command == 8 || command == 9)	/* II read sector */
		{
			go_to_sector(disk_reg, track);

			if (verbose_disk) fprintf(stderr, "ii read sector at %d(%d):%d multiple: %d:\n", track, disk_reg[0x09], disk_reg[0x0a], m);
			if (verbose_disk) debug_show_disk_pos();

			if (read(disk_fh, io_buffer, 512) != 512)
			{
				if (verbose_disk) fprintf(stderr, "short read reading from file\n");
			}
			if (verbose_disk) fprintf(stderr, "sector %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
						  io_buffer[0], io_buffer[1], io_buffer[2], io_buffer[3], io_buffer[4], io_buffer[5], io_buffer[6], io_buffer[7],
						  io_buffer[8], io_buffer[9], io_buffer[10], io_buffer[11], io_buffer[12], io_buffer[13], io_buffer[14], io_buffer[15]);
			io_buffer_index = 0;
			io_buffer_left  = 512;

			rws = st_read;
			tc = type_ii;
			//verbose_z80 = 1;
			flags |= T2_BUSY | T2_DRQ;
			busy_cnt = 5;
		}
		else if (command == 10 || command == 11)/* II write sector */
		{
			if (go_to_sector(disk_reg, track) == 0)
			{
				rws = st_write;
			}

			if (verbose_disk) fprintf(stderr, "ii write sector at %d(%d):%d multiple: %d:\n", track, disk_reg[0x09], disk_reg[0x0a], m);
			if (verbose_disk) debug_show_disk_pos();

			io_buffer_index = 0;
			io_buffer_left  = 512;

			tc = type_ii;
			write_mode = wm_sector;

			flags |= T2_BUSY | T2_DRQ;
			busy_cnt = 5;
		}
		else if (command == 12)			/* III read address */
		{
			if (verbose_disk) fprintf(stderr, "iii read address\n");
			tc = type_iii;

			io_buffer[0] = track;
			io_buffer[1] = disk_reg[0x0c] & 10 ? 1 : 0;
			io_buffer[2] = disk_reg[0x0a];
			io_buffer[3] = 0x01;	/* sector length, 0x01 might be incorrect */
			io_buffer[4] = 0x00;	/* CRC1 */
			io_buffer[5] = 0x00;	/* CRC2 */
			io_buffer_index = 0;
			io_buffer_left = 6;
			rws = st_read;
			tc = type_iii;
			flags |= T2_BUSY | T2_DRQ;	// fixme: should be T3_
			busy_cnt = 5;
		}
		else if (command == 14)			/* III read track */
		{
			if (verbose_disk) fprintf(stderr, "iii read track\n");
			tc = type_iii;
			if (go_to_track(track) == 0)
			{
				// fixme
			}
			flags |= T2_BUSY;
			busy_cnt = 5;
		}
		else if (command == 15)			/* III write track */
		{
			if (verbose_disk) fprintf(stderr, "iii write track\n");
			tc = type_iii;
			write_mode = wm_track;
			io_buffer_index = 0;
			io_buffer_left = (115 + 512) * 9 + 146 + 10; /* 10 == track end, 0x4e bytes */
			if (go_to_track(track) == 0)
			{
				rws = st_write;
			}
			flags |= T2_BUSY | T2_DRQ;
			busy_cnt = 5;
		}
		else if (command == 13)			/* IV force interrupt */
		{
			if (verbose_disk) fprintf(stderr, "iv force interrupt %d\n", i);
			tc = type_iv;

			io_buffer_index = io_buffer_left = 0;
			rws = st_idle;
		}
	}
	else if (address == 0x0b)	/* data register */
	{
		if (rws == st_idle)
		{
			if (verbose_disk) fprintf(stderr, "Setting up data register for seek to %d\n", value);
		}
		else if (rws == st_write)
		{
			if (verbose_disk && io_buffer_index == 0) fprintf(stderr, "write data %02x at index %d\n", value, io_buffer_index);

			io_buffer[io_buffer_index++] = value;
			io_buffer_left--;

			if (io_buffer_left == 0)
			{
				if (write_mode == wm_sector)
				{
					if (verbose_disk)
					{
						fprintf(stderr, "write sector finished after %d bytes, will write at:\n", io_buffer_index);
						debug_show_disk_pos();
					}

					if (write(disk_fh, io_buffer, 512) != 512)
					{
						if (verbose_disk) fprintf(stderr, "short write writing to file\n");
					}
				}
				else if (write_mode == wm_track)
				{
					int offset;
					int loop;

					if (verbose_disk)
					{
						fprintf(stderr, "write track finished after %d bytes, will write at:\n", io_buffer_index);
						debug_show_disk_pos();
					}

					if (lseek(disk_fh, file_offset(get_side(disk_reg), track, 1), SEEK_SET)  == -1)
						error_exit("seek error");

					/* FIXME: write track */
					offset = 80 + 12 + 3 + 1 + 26/*50<-wrong*/;	/* track header */
					for(loop=0; loop<9; loop++)
					{
						offset += 12 + 3 + 6 + 22;	/* sector id field */
						offset += 12 + 3 + 1;		/* sector data field (1) */

						if (write(disk_fh, &io_buffer[offset], 512) != 512)
						{
							if (verbose_disk) fprintf(stderr, "short write writing track to file\n");
						}
						offset += 512;

						offset += 1 + 54;		/* sector data field (2) */
					}
					if (verbose_disk) fprintf(stderr, "index: %d offset: %d\n", io_buffer_index, offset);
					//{FILE *fh = fopen("track.dmp", "w"); fwrite(io_buffer, 1, io_buffer_index, fh); fclose(fh); exit(0); }
					if (verbose_disk) fprintf(stderr, "0x4e: %02x\n", io_buffer[offset]);
				}

				io_buffer_index = 0;
				io_buffer_left = 0;

				rws = st_idle;

				flags &= ~(T2_DRQ | T2_BUSY);
			}
			else
			{
				flags |= T2_DRQ;
			}
		}
		else
		{
			if (verbose_disk) fprintf(stderr, "write data-register not in write/idle mode (offset %d)\n", io_buffer_index);
		}
	}
	else if (address == 0x09)			/* track */
	{
		if (verbose_disk) fprintf(stderr, "set track to %d (%d)\n", value, disk_reg[address]);
		/* handled by disk_reg[address] = value */
	}
	else if (address == 0x0a)			/* sector */
	{
		if (verbose_disk) fprintf(stderr, "set sector to %d (%d)\n", value, disk_reg[address]);
		/* handled by disk_reg[address] = value */
	}
	else if (address == 0x0c)			/* side */
	{
		/* handled by disk_reg[address] = value */
		if (value & 0x04) /* RESET */
		{
			if (verbose_disk) fprintf(stderr, "resetting controller\n");
			track = 0;
			memset(disk_reg, 0x00, sizeof(disk_reg));
		}

		if (verbose_disk) fprintf(stderr, "side: %d, drive: %d\n", value & 0x10 ? 1 : 0, value & 0x03);
	}
	else if (address == 0x0d)			/* disk motor */
	{
		if (verbose_disk) fprintf(stderr, "set motor to %s (%d / %d)\n", value & 1?"off":"on", value, value & 1);
	}
	else
	{
		if (verbose_disk) fprintf(stderr, "unhandled disk write: %04x %02x\n", address, value);
	}

	mvwprintw(stdscr, 10, 0, "regW %x:%02x %d|%02d|%d %04d   ", address & 0xf, value, value & 0x10 ? 1 : 0, track, disk_reg[0x0a], io_buffer_index);
}
