/* msx-emul is (C) 2008-2009 by folkert@vanheusden.com
 * released under GPL
 */
static char *rcsId = "$Id: keyboard.c,v 1.17 2009/02/18 14:16:50 folkert Exp folkert $";

extern char verbose_cpu;
#include <z80ex/z80ex.h>
#include <z80ex/z80ex_dasm.h>
#include <ncurses.h>
#include <SDL/SDL.h>
#include <sndfile.h>

#include "gen.h"
#include "msx.h"
#include "sound.h"
#include "disk.h"
#include "video.h"

#define KBUF_SIZE       8

char dump = 0;

typedef enum { mod_none, mod_shift, mod_ctrl } mod_t;

typedef struct
{
	int row1, bit1, ignore1;
	int row2, bit2, ignore2;
	int real_char;
	mod_t mod;
} keymapper_t;

keymapper_t map[] = {
	{ 0, 0, 0, -1, -1, 1, SDLK_0, mod_none },
	{ 0, 0, 0,  6,  0, 0, SDLK_0, mod_shift },	/* ) */
	{ 0, 1, 0, -1, -1, 1, SDLK_1, mod_none },
	{ 0, 1, 0,  6,  0, 0, SDLK_1, mod_shift },	/* ! */
	{ 0, 2, 0, -1, -1, 1, SDLK_2, mod_none },
	{ 0, 2, 0,  6,  0, 0, SDLK_2, mod_shift },	/* @ */
	{ 0, 3, 0, -1, -1, 1, SDLK_3, mod_none },
	{ 0, 3, 0,  6,  0, 0, SDLK_3, mod_shift },	/* # */
	{ 0, 4, 0, -1, -1, 1, SDLK_4, mod_none },
	{ 0, 4, 0,  6,  0, 0, SDLK_4, mod_shift },	/* $ */
	{ 0, 5, 0, -1, -1, 1, SDLK_5, mod_none },
	{ 0, 5, 0,  6,  0, 0, SDLK_5, mod_shift },	/* % */
	{ 0, 6, 0, -1, -1, 1, SDLK_6, mod_none },
	{ 0, 6, 0,  6,  0, 0, SDLK_6, mod_shift },	/* ^ */
	{ 0, 7, 0, -1, -1, 1, SDLK_7, mod_none },
	{ 0, 7, 0,  6,  0, 0, SDLK_7, mod_shift },	/* & */
	{ 1, 0, 0, -1, -1, 1, SDLK_8, mod_none },
	{ 1, 0, 0,  6,  0, 0, SDLK_8, mod_shift },	/* * */
	{ 1, 1, 0, -1, -1, 1, SDLK_9, mod_none },
	{ 1, 1, 0,  6,  0, 0, SDLK_9, mod_shift },	/* ( */
	{ 1, 2, 0, -1, -1, 1, SDLK_MINUS, mod_none },	/* - */
	{ 1, 2, 0,  6,  0, 0, SDLK_MINUS, mod_shift },	/* _ */
	{ 1, 3, 0, -1, -1, 1, SDLK_EQUALS, mod_none },	/* = */
	{ 1, 3, 0,  6,  0, 0, SDLK_EQUALS, mod_shift },	/* + */
	{ 1, 4, 0, -1, -1, 1, SDLK_BACKSLASH, mod_none },	/* \ */
	{ 1, 4, 0,  6,  0, 0, SDLK_BACKSLASH, mod_none },	/* | */
	{ 1, 5, 0, -1, -1, 1, SDLK_LEFTBRACKET, mod_none },	/* [ */
	{ 1, 5, 0,  6,  0, 0, SDLK_LEFTBRACKET, mod_shift },	/* { */
	{ 1, 6, 0, -1, -1, 1, SDLK_RIGHTBRACKET, mod_none },	/* ] */
	{ 1, 6, 0,  6,  0, 0, SDLK_RIGHTBRACKET, mod_shift },	/* } */
	{ 1, 7, 0, -1, -1, 1, SDLK_SEMICOLON, mod_none },	/* ; */
	{ 1, 7, 0,  6,  0, 0, SDLK_SEMICOLON, mod_shift },	/* : */
	{ 2, 0, 0, -1, -1, 1, SDLK_QUOTE, mod_none },		/* ' */
	{ 2, 0, 0,  6,  0, 0, SDLK_QUOTE, mod_shift },		/* " */
	{ 2, 1, 0, -1, -1, 1, SDLK_BACKQUOTE, mod_none },	/* ` */
	{ 2, 1, 0,  6,  0, 0, SDLK_BACKQUOTE, mod_shift },	/* ~ */
	{ 2, 2, 0, -1, -1, 1, SDLK_COMMA, mod_none },	/* , */
	{ 2, 2, 0,  6,  0, 0, SDLK_COMMA, mod_shift },	/* < */
	{ 2, 3, 0, -1, -1, 1, SDLK_PERIOD, mod_none },	/* . */
	{ 2, 3, 0,  6,  0, 0, SDLK_PERIOD, mod_shift },	/* > */
	{ 2, 4, 0, -1, -1, 1, SDLK_SLASH, mod_none },	/* / */
	{ 2, 4, 0,  6,  0, 0, SDLK_SLASH, mod_shift },	/* ? */
	/* 2, 5, 0 is not mapped */
	{ 2, 6, 0, -1, -1, 1, SDLK_a, mod_none },
	{ 2, 6, 0,  6,  0, 0, SDLK_a, mod_shift },
	{ 2, 7, 0, -1, -1, 1, SDLK_b, mod_none },
	{ 2, 7, 0,  6,  0, 0, SDLK_b, mod_shift },
	{ 3, 0, 0, -1, -1, 1, SDLK_c, mod_none },
	{ 3, 0, 0,  6,  0, 0, SDLK_c, mod_shift },
	{ 3, 1, 0, -1, -1, 1, SDLK_d, mod_none },
	{ 3, 1, 0,  6,  0, 0, SDLK_d, mod_shift },
	{ 3, 2, 0, -1, -1, 1, SDLK_e, mod_none },
	{ 3, 2, 0,  6,  0, 0, SDLK_e, mod_shift },
	{ 3, 3, 0, -1, -1, 1, SDLK_f, mod_none },
	{ 3, 3, 0,  6,  0, 0, SDLK_f, mod_shift },
	{ 3, 4, 0, -1, -1, 1, SDLK_g, mod_none },
	{ 3, 4, 0,  6,  0, 0, SDLK_g, mod_shift },
	{ 3, 5, 0, -1, -1, 1, SDLK_h, mod_none },
	{ 3, 5, 0,  6,  0, 0, SDLK_h, mod_shift },
	{ 3, 6, 0, -1, -1, 1, SDLK_i, mod_none },
	{ 3, 6, 0,  6,  0, 0, SDLK_i, mod_shift },
	{ 3, 7, 0, -1, -1, 1, SDLK_j, mod_none },
	{ 3, 7, 0,  6,  0, 0, SDLK_j, mod_shift },
	{ 4, 0, 0, -1, -1, 1, SDLK_k, mod_none },
	{ 4, 0, 0,  6,  0, 0, SDLK_k, mod_shift },
	{ 4, 1, 0, -1, -1, 1, SDLK_l, mod_none },
	{ 4, 1, 0,  6,  0, 0, SDLK_l, mod_shift },
	{ 4, 2, 0, -1, -1, 1, SDLK_m, mod_none },
	{ 4, 2, 0,  6,  0, 0, SDLK_m, mod_shift },
	{ 4, 3, 0, -1, -1, 1, SDLK_n, mod_none },
	{ 4, 3, 0,  6,  0, 0, SDLK_n, mod_shift },
	{ 4, 4, 0, -1, -1, 1, SDLK_o, mod_none },
	{ 4, 4, 0,  6,  0, 0, SDLK_o, mod_shift },
	{ 4, 5, 0, -1, -1, 1, SDLK_p, mod_none },
	{ 4, 5, 0,  6,  0, 0, SDLK_p, mod_shift },
	{ 4, 6, 0, -1, -1, 1, SDLK_q, mod_none },
	{ 4, 6, 0,  6,  0, 0, SDLK_q, mod_shift },
	{ 4, 7, 0, -1, -1, 1, SDLK_r, mod_none },
	{ 4, 7, 0,  6,  0, 0, SDLK_r, mod_shift },
	{ 5, 0, 0, -1, -1, 1, SDLK_s, mod_none },
	{ 5, 0, 0,  6,  0, 0, SDLK_s, mod_shift },
	{ 5, 1, 0, -1, -1, 1, SDLK_t, mod_none },
	{ 5, 1, 0,  6,  0, 0, SDLK_t, mod_shift },
	{ 5, 2, 0, -1, -1, 1, SDLK_u, mod_none },
	{ 5, 2, 0,  6,  0, 0, SDLK_u, mod_shift },
	{ 5, 3, 0, -1, -1, 1, SDLK_v, mod_none },
	{ 5, 3, 0,  6,  0, 0, SDLK_v, mod_shift },
	{ 5, 4, 0, -1, -1, 1, SDLK_w, mod_none },
	{ 5, 4, 0,  6,  0, 0, SDLK_w, mod_shift },
	{ 5, 5, 0, -1, -1, 1, SDLK_x, mod_none },
	{ 5, 5, 0,  6,  0, 0, SDLK_x, mod_shift },
	{ 5, 6, 0, -1, -1, 1, SDLK_y, mod_none },
	{ 5, 6, 0,  6,  0, 0, SDLK_y, mod_shift },
	{ 5, 7, 0, -1, -1, 1, SDLK_z, mod_none },
	{ 5, 7, 0,  6,  0, 0, SDLK_z, mod_shift },
	{ 6, 0, 0, -1, -1, 1, SDLK_LSHIFT, mod_none },
	{ 6, 0, 0, -1, -1, 1, SDLK_RSHIFT, mod_none },
	{ 6, 1, 0, -1, -1, 1, SDLK_RCTRL, mod_none },
	{ 6, 1, 0, -1, -1, 1, SDLK_LCTRL, mod_none },
	/* 6, 2: graph */
	/* 6, 3: cap   */
	/* 6, 4: code */
	{ 6, 5, 0, -1, -1, 1, SDLK_F1, mod_none },
	{ 6, 5, 0,  6,  0, 0, SDLK_F6, mod_none },
	{ 6, 6, 0, -1, -1, 1, SDLK_F2, mod_none },
	{ 6, 6, 0,  6,  0, 0, SDLK_F7, mod_none },
	{ 6, 7, 0, -1, -1, 1, SDLK_F3, mod_none },
	{ 6, 7, 0,  6,  0, 0, SDLK_F8, mod_none },
	{ 7, 0, 0, -1, -1, 1, SDLK_F4, mod_none },
	{ 7, 0, 0,  6,  0, 0, SDLK_F9, mod_none },
	{ 7, 1, 0, -1, -1, 1, SDLK_F5, mod_none },
	{ 7, 1, 0,  6,  0, 0, SDLK_F10, mod_none },
	{ 7, 2, 0, -1, -1, 1, SDLK_ESCAPE, mod_none },
	{ 7, 3, 0, -1, -1, 1, 9, mod_none },
	{ 7, 4, 0, -1, -1, 1, SDLK_F12, mod_none },	/* STOP: F12 */
	{ 7, 5, 0, -1, -1, 1, 8, mod_none },
	{ 7, 6, 0, -1, -1, 1, SDLK_F11, mod_none },	/* SELECT: F11 */
	{ 7, 7, 0, -1, -1, 1, SDLK_RETURN, mod_none },
	{ 8, 0, 0, -1, -1, 1, SDLK_SPACE, mod_none },
	{ 8, 1, 0, -1, -1, 1, SDLK_HOME, mod_none },
	{ 8, 2, 0, -1, -1, 1, SDLK_INSERT, mod_none },
	{ 8, 3, 0, -1, -1, 1, SDLK_DELETE, mod_none },
	{ 8, 4, 0, -1, -1, 1, SDLK_LEFT, mod_none },
	{ 8, 5, 0, -1, -1, 1, SDLK_UP, mod_none },
	{ 8, 6, 0, -1, -1, 1, SDLK_DOWN, mod_none },
	{ 8, 7, 0, -1, -1, 1, SDLK_RIGHT, mod_none },
	{ -1, -1, 1, -1, -1, -1, -1 }
};
int map_size = sizeof(map) / sizeof(keymapper_t);

keymapper_t keybuffer[KBUF_SIZE];
int keybuffer_in = 0;
unsigned char rows[16] = { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 };

int keys_dropped = 0, keys_processed = 0;

unsigned char read_keyboard(int row)
{
	mvwprintw(stdscr, 6, 31, "row: %d ", row);

	if (verbose)
	{
		if (rows[row] != 255)
			fprintf(stderr, "checked row %d: %d\n", row, rows[row]);
	}

	return rows[row];
}

void check_keyboard(void)
{
	SDL_Event event;

	if (SDL_PollEvent(&event))
	{
		int loop;

		if (event.type == SDL_KEYDOWN || event.type == SDL_KEYUP)
		{
			int loop, scanned = -1;
			int typed = event.key.keysym.sym;
			int mod   = event.key.keysym.mod;
			int key_shift = ((mod & KMOD_LSHIFT) != 0) | ((mod & KMOD_RSHIFT) != 0);
			int key_ctrl  = ((mod & KMOD_LCTRL ) != 0) | ((mod & KMOD_RCTRL ) != 0);

			fprintf(stderr, "sym: %d, mod: %d, shift: %d, ctrl: %d\n", typed, mod, key_shift, key_ctrl);

			for(loop=0; loop<map_size; loop++)
			{
				if (map[loop].ignore1 == 0 &&
				    map[loop].real_char == typed)
				{
					if (key_shift)
					{
						if (map[loop].ignore2 == 0 && map[loop].mod == mod_shift)
						{
							scanned = loop;
							break;
						}
					}
					else if (key_ctrl)
					{
						if (map[loop].ignore2 == 0 && map[loop].mod == mod_ctrl)
						{
							scanned = loop;
							break;
						}
					}
					else
					{
						if (map[loop].ignore2)
						{
							scanned = loop;
							break;
						}
					}
				}
			}

			if (scanned != -1)
			{
				if (verbose) fprintf(stderr, "key %s: |%c| row1: %d, bit1: %d, row2: %d, bit2: %d\n",
						event.type == SDL_KEYDOWN ? "down":"  up",
						map[scanned].real_char,
						map[scanned].row1, map[scanned].bit1,
						map[scanned].row2, map[scanned].bit2);

				if (event.type == SDL_KEYDOWN)
				{
					if (map[scanned].ignore1 == 0)
						rows[map[scanned].row1] &= ~(1 << map[scanned].bit1);

					if (map[scanned].ignore2 == 0)
						rows[map[scanned].row2] &= ~(1 << map[scanned].bit2);
				}
				else
				{
					if (map[scanned].ignore1 == 0)
						rows[map[scanned].row1] |= (1 << map[scanned].bit1);

					if (map[scanned].ignore2 == 0)
						rows[map[scanned].row2] |= (1 << map[scanned].bit2);
				}

				mvwprintw(stdscr, 6, 19, "last: |%c|", typed);
			}
		}
		else if (event.type == SDL_QUIT)
		{
			if (sndfile_out)
			{
				sf_write_sync(sndfile_out);
				fprintf(stderr, "Closing wav-file\n");
				sf_close(sndfile_out);
			}

			FILE *fh = fopen("vram.dmp", "wb");
			fwrite(vdp_mem, 128*1024, 1, fh);
			fclose(fh);

			if (disk_fh != -1)
				close(disk_fh);

			endwin();
			exit(0);
		}

		for(loop=0; loop<16; loop++)
			mvwprintw(stdscr, 11 + loop, 82, "%02x", rows[loop]);
	}
}
