/* msx-emul is (C) 2008-2009 by folkert@vanheusden.com
 * released under GPL
 */
static char *rcsId = "$Id: memory.c,v 1.15 2009/02/18 14:17:04 folkert Exp folkert $";

#include <z80ex/z80ex.h>
#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <SDL/SDL.h>
#include <sndfile.h>

#include "error.h"
#include "msx.h"
#include "ports.h"
#include "disk.h"
#include "video.h"
#include "sound.h"
#include "memory.h"

extern char verbose_mem;
extern char verbose_mem_conf;
extern char verbose_mem_err;
extern char verbose_mem_mapper;
extern char verbose_mem_megarom;

/* generic slots */
unsigned char *memory[4][4][4];
char rom[4][4][4];
char has_subpages[4];
unsigned char sub_page_selection[4] = { 0x00, 0x00, 0x00, 0x00 };

/* MEGAROM */
megarom_t *megaroms = NULL;
int n_megaroms = 0;

/* memory mapper */
mapper_t *mappers = NULL;
int n_mappers = 0;
unsigned char mapper_bank_selection[4] = { 0, 1, 2, 3 };

/* memory mapper */
Z80EX_BYTE mapper_read(int slot, int sub_slot, Z80EX_WORD address)
{
	int loop;

	for(loop=0; loop<n_mappers; loop++)
	{
		if (mappers[loop].mapper_slot == slot && mappers[loop].mapper_sub_slot == sub_slot)
		{
			Z80EX_BYTE value;
			int bank = address >> 14;
			int offset = address & 0x3fff;

			value = (mappers[loop].mapper)[mapper_bank_selection[bank] * 0x4000 + offset];

			if (verbose_mem_mapper) fprintf(stderr, "read from page %d (mapped to bank %d) at offset %04x: %02x\n", bank, mapper_bank_selection[bank], offset, value);

			return value;
		}
	}

	error_exit("mapper_read: %d:%d not found", slot, sub_slot);

	return -1;
}

void mapper_write(int slot, int sub_slot, Z80EX_WORD address, Z80EX_BYTE value)
{
	int loop;

	for(loop=0; loop<n_mappers; loop++)
	{
		if (mappers[loop].mapper_slot == slot && mappers[loop].mapper_sub_slot == sub_slot)
		{
			int bank = address >> 14;
			int offset = address & 0x3fff;

			(mappers[loop].mapper)[mapper_bank_selection[bank] * 0x4000 + offset] = value;

			if (verbose_mem_mapper) fprintf(stderr, "write to page %d (mapped to bank %d) at offset %04x: %02x\n", bank, mapper_bank_selection[bank], offset, value);

			return;
		}
	}

	error_exit("mapper_write: mapper %d:%d not found", slot, sub_slot);
}

/* MEGAROM code */
Z80EX_BYTE megarom_read(int slot, int sub_slot, Z80EX_WORD address)
{
	int loop;

	for(loop=0; loop<n_megaroms; loop++)
	{
		if (megaroms[loop].megarom_slot == slot && megaroms[loop].megarom_sub_slot == sub_slot)
		{
			switch(megaroms[loop].megarom_mapper_type)
			{
				case rom_megarom_konami4:		/* only bankswitch addresses differ from 5 */
				case rom_megarom_konami5:
					if (address >= 0x4000 && address <= 0xbfff)
					{
						int bank = (address >> 13) - 2;
						int offset = address & 0x1fff;

						int rom_offset = megaroms[loop].megarom_bank_selection[bank] * 0x2000 + offset;

						if (rom_offset >= megaroms[loop].size)
							fprintf(stderr, "Reading past end of megarom\n");

						return (megaroms[loop].megarom)[rom_offset];
					}

					break;

				case rom_megarom_msxdos2:
					if (address >= 0x4000 && address <= 0x7fff)
					{
						return (megaroms[loop].megarom)[megaroms[loop].megarom_bank_selection[0] * 0x4000 + address - 0x4000];
					}
					break;

				case rom_ascii_16kb:
					if (address >= 0x4000 && address <= 0xbfff)
					{
						int bank = (address - 0x4000) / 0x4000;
						int offset = address & 0x3fff;

						int rom_offset = megaroms[loop].megarom_bank_selection[bank] * 0x4000 + offset;

						if (rom_offset >= megaroms[loop].size)
							fprintf(stderr, "Reading past end of megarom: %x, bank %d, offset %04x\n", rom_offset, bank, offset);

						return megaroms[loop].megarom[rom_offset];
					}
					break;

				default:
					error_exit("read: megarom type %d not supported", megaroms[loop].megarom_mapper_type);
			}

			return -1;
		}
	}

	if (verbose_mem_megarom) fprintf(stderr, "megarom_read out of range %04x\n", address);

	return -1;
}

void megarom_write(int slot, int sub_slot, Z80EX_WORD address, Z80EX_BYTE value)
{
	int loop;

	for(loop=0; loop<n_megaroms; loop++)
	{
		if (megaroms[loop].megarom_slot == slot && megaroms[loop].megarom_sub_slot == sub_slot)
		{
			switch(megaroms[loop].megarom_mapper_type)
			{
				case rom_megarom_konami4:	/* range differ, 4 has static bank 0 */
				case rom_megarom_konami5:
					{
						int bank = (address >> 13) - 2;
						int offset = address & 0x1fff;
						char ok = 0;

						if (megaroms[loop].megarom_mapper_type == rom_megarom_konami5)
						{
							ok = bank >= 0 && bank < 4 && offset >= 0x1000 && offset <= 0x17ff;
							if (address >= 0x9000 && address <= 0x97ff && (value & 0x3f) == 0x3f)
							{
								fprintf(stderr, "ENABLE SCC\n");
								megaroms[loop].scc_selected = 1;
								scc_enabled = 1;
							}
						}
						else if (megaroms[loop].megarom_mapper_type == rom_megarom_konami4)
							ok = bank >= 1 && bank < 4;

						if (ok)
						{
							value &= 63; /* only lower 6 bits thus 512KB */
							megaroms[loop].megarom_bank_selection[bank] = value;

							if (verbose_mem_megarom) fprintf(stderr, "megarom (konami4/5) set bank %d to %d\n", bank, value);

							mvwprintw(stdscr, 9, 0, "megarom: %02x %02x %02x %02x",
									megaroms[loop].megarom_bank_selection[0], megaroms[loop].megarom_bank_selection[1],
									megaroms[loop].megarom_bank_selection[2], megaroms[loop].megarom_bank_selection[3]);
						}
						else if (verbose_mem_megarom)
						{
							fprintf(stderr, "megarom_write (konami4/5) out of range %04x - %02x: bank %d, offset %04x\n",
									address, value, bank, offset);
						}

						if (megaroms[loop].scc_selected && address >= 0x9800 && address <= 0xafff)
						{
							scc_regs[address & 0xff] = value;
							scc_sound();
						}
					}
					return;	/* !!! */

				case rom_megarom_msxdos2:
					if (address == 0x6000 || address == 0x7ffe)
					{
						value &=3;
						megaroms[loop].megarom_bank_selection[0] = value;

						if (verbose_mem_megarom) fprintf(stderr, "megarom (msxdos2) set to %d\n", value);
					}
					else
					{
						if (verbose_mem_megarom) fprintf(stderr, "megarom_write (msxdos2) out of range %04x - %02x\n", address, value);
					}
					return;	/* !!! */

				case rom_ascii_16kb:
					if (address >= 0x6000 && address <= 0x67ff)
						megaroms[loop].megarom_bank_selection[0] = value;
					else if (address >= 0x7000 && address <= 0x77ff)
						megaroms[loop].megarom_bank_selection[1] = value;
					return;

				default:
					error_exit("write: megarom type %d not supported", megaroms[loop].megarom_mapper_type);
			}
		}
	}

	error_exit("megarom_write: mapper %d:%d not found", slot, sub_slot);
}
/* ********************** */

inline void split_addr(Z80EX_WORD addr, int *page, int *slot, int *sub_page, int *mem)
{
	int shift;
	*page = addr >> 14;
	shift = *page * 2;
	*slot = (ports[0xa8] >> shift) & 0x03;
	if (has_subpages[*slot])
		*sub_page = (sub_page_selection[*slot] >> shift) & 0x03;
	else
		*sub_page = 0;
	*mem  = addr & 0x3fff;
}

Z80EX_BYTE read_mem(Z80EX_WORD addr, char verbose_mem)
{
	unsigned char *p;
	int page, slot, sub_page, mem;

	split_addr(addr, &page, &slot, &sub_page, &mem);

	if (page == 3 && mem == 0x3fff && has_subpages[slot])
	{
		if (verbose_mem_conf) printf("read subpage %02x\n", sub_page_selection[slot]);
		return 255 - sub_page_selection[slot];
	}
	else if (rom[slot][page][sub_page] == 2)
	{
		return mapper_read(slot, sub_page, addr);
	}
	else if (rom[slot][page][sub_page] == 3)
	{
		return megarom_read(slot, sub_page, addr);
	}
	else if (slot == disk_slot && mem >= 0x3ff8 && page == 1 && sub_page == disk_sub_slot)
	{
		return disk_memorymapped_io_read(mem - 0x3ff0);
	}

	p = memory[slot][page][sub_page];

	if (p)
		return p[mem];

	return 0xee; /* reading invalid memory */
}

Z80EX_BYTE read_mem_dis(Z80EX_WORD addr, void *user_data)
{
	return read_mem(addr, 0);
}

Z80EX_BYTE read_mem_cb(Z80EX_CONTEXT *cpu, Z80EX_WORD addr, int m1_state, void *user_data)
{
	return read_mem(addr, verbose_mem);
}

void write_mem_cb(Z80EX_CONTEXT *cpu, Z80EX_WORD addr, Z80EX_BYTE value, void *user_data)
{
	int page, slot, sub_page, mem;
	unsigned char *p;

	split_addr(addr, &page, &slot, &sub_page, &mem);

if (addr == 0xfafc)
fprintf(stderr, "0xfafc write: %02x\n", value);

	if (page == 3 && mem == 0x3fff && has_subpages[slot])
	{
		if (verbose_mem_conf) fprintf(stderr, "write subpage %02x\n", value);
		sub_page_selection[slot] = value;
		mvwprintw(stdscr, 2, 40, "%02x %02x %02x %02x", sub_page_selection[0], sub_page_selection[1], sub_page_selection[2], sub_page_selection[3]);
	}
	else if (rom[slot][page][sub_page] == 2)
	{
		mapper_write(slot, sub_page, addr, value);
	}
	else if (rom[slot][page][sub_page] == 3)
	{
		megarom_write(slot, sub_page, addr, value);
	}
	else if (slot == disk_slot && mem >= 0x3ff8 && page == 1 && sub_page == disk_sub_slot)
	{
		// fprintf(stderr, "AF: %04x, BC: %04x, DE: %04x, HL: %04x, PC: %04x, SP: %04x, IX: %04x, IY: %04x\n", z80ex_get_reg(cpu, regAF), z80ex_get_reg(cpu, regBC), z80ex_get_reg(cpu, regDE), z80ex_get_reg(cpu, regHL), z80ex_get_reg(cpu, regPC), z80ex_get_reg(cpu, regSP), z80ex_get_reg(cpu, regIX), z80ex_get_reg(cpu, regIY));
		disk_memorymapped_io_write(mem - 0x3ff0, value);
	}
	else
	{
		p = memory[slot][page][sub_page];

		if (rom[slot][page][sub_page] == 0 && p != NULL)	/* ram */
		{
			p[mem] = value;
		}
		else if (rom[slot][page][sub_page] == 2) /* mapper */
		{
			/* handled above */
			error_exit("write_mem_cb: internal error");
		}
		else if (rom[slot][page][sub_page] == 1) /* generic rom */
		{
			static long long int rom_writes = 0;

			rom_writes++;
			mvwprintw(stdscr, 1, 40, "ROMwrites: %lld|%04x", rom_writes, addr);

			if (verbose_mem_err) fprintf(stderr, "Writing to ROM slot %d/page %d/subpage %d, address %04x (%04x), value: %02x\n", slot, page, sub_page, addr, mem, value);
		}
		else if (verbose_mem_err)
			fprintf(stderr, "WRITE no mem/rom: slot %d/page %d/subpage %d, address %04x, value: %02x, 0xa8: %02x\n", slot, page, sub_page, mem, value, ports[0xa8]);
	}
}

int load_rom_page(FILE *fh, int slot, int page, int sub_page)
{
	int rc;

	memory[slot][page][sub_page] = (unsigned char *)malloc(16384);

	rc = fread(memory[slot][page][sub_page], 1, 16384, fh);
	if (rc < 16384)
	{
		free(memory[slot][page][sub_page]);
		memory[slot][page][sub_page] = NULL;
		return -1;
	}

	fprintf(stderr, " reading %d:%d page %d (= %04x)\n", slot, sub_page, page, page * 0x4000);

	rom[slot][page][sub_page] = 1;

	return 0;
}

void init_memory(roms_t *roms, int n_roms)
{
	FILE *fh;
	int page, slot, sub_page;
	int rom_index, mr_index;

	if (msx_level == 1)
		vdp_mem = (unsigned char *)malloc(16384);	/* VDP ram */
	else
		vdp_mem = (unsigned char *)malloc(128 * 1024);	/* VDP ram */

	for(rom_index=0; rom_index<n_roms; rom_index++)
	{
		int page, size;

		fh = fopen(roms[rom_index].file, "r");
		if (!fh)
			error_exit("failed loading %s", roms[rom_index].file);

		fprintf(stderr, "Load %s to %d:%d\n", roms[rom_index].file, roms[rom_index].slot, roms[rom_index].subslot);

		switch(roms[rom_index].type)
		{
			case rom_bios:
				fprintf(stderr, " is BIOS, at offset 0x0000\n");
				page = 0;
				for(;;)
				{
					if (load_rom_page(fh, roms[rom_index].slot, page, roms[rom_index].subslot) == -1)
						break;
					page++;
				}
				break;

			case rom_generic:
				page = 1;
				fprintf(stderr, " is generic ROM, loading to page %d (offset %04x)\n", page, page * 0x4000);
				for(;;)
				{
					if (load_rom_page(fh, roms[rom_index].slot, page, roms[rom_index].subslot) == -1)
						break;
					page++;
				}
				break;

			case rom_diskrom:
				fprintf(stderr, " is disk-ROM, loading in page 1\n");
				if (load_rom_page(fh, roms[rom_index].slot, 1, roms[rom_index].subslot) == 0)
				{
					disk_slot = roms[rom_index].slot;
					disk_sub_slot = roms[rom_index].subslot;
					disk_file = roms[rom_index].parameter;
				}
				else
					error_exit("failed loading disk-ROM");
				break;

			case rom_megarom_konami4: /* konami4 */
			case rom_megarom_konami5: /* konami5 */
			case rom_megarom_msxdos2: /* MSX-DOS 2 */
			case rom_ascii_16kb:
				fprintf(stderr, " is MegaROM (%d)\n", roms[rom_index].type);
				for(mr_index=0; mr_index<n_megaroms; mr_index++)
				{
					if (roms[rom_index].slot == megaroms[mr_index].megarom_slot &&
							roms[rom_index].subslot == megaroms[mr_index].megarom_sub_slot)
					{
						fseek(fh, 0, SEEK_END);
						megaroms[mr_index].size = ftell(fh);
						fprintf(stderr, " megarom size: %d\n", megaroms[mr_index].size);
						fseek(fh, 0, SEEK_SET);
						megaroms[mr_index].megarom = (unsigned char *)malloc(megaroms[mr_index].size);
						fread(megaroms[mr_index].megarom, 1, megaroms[mr_index].size, fh);
						break;
					}
				}
				break;

			case rom_msx2exp:
				fprintf(stderr, " is MSX2 sub-rom\n");
				if (load_rom_page(fh, roms[rom_index].slot, 0, roms[rom_index].subslot) == -1)
					error_exit("failed loading MSX2 sub-rom");
				break;

			default:
				error_exit("unhandled rom-type");
		}

		has_subpages[roms[rom_index].slot] |= roms[rom_index].subslot > 0;

		fclose(fh);
	}

	for(page = 3; page >=0; page--)
	{
		fprintf(stderr, "p%02x | ", page);

		for(slot = 0; slot < 4; slot++)
		{
			for(sub_page=0; sub_page<4; sub_page++)
			{
				if (rom[slot][page][sub_page] == 3 || rom[slot][page][sub_page] == 2)
					fprintf(stderr, "1");
				else
					fprintf(stderr, "%1x", memory[slot][page][sub_page] ? 1:0);
			}

			fprintf(stderr, " ");
		}

		fprintf(stderr, "\n");
	}
	fprintf(stderr, "      ");
	for(slot=0; slot<4; slot++)
		fprintf(stderr, " %1x-%1x ", slot, has_subpages[slot]);
	fprintf(stderr, "\n");
}
