/* msx-emul is (C) 2008-2010 by folkert@vanheusden.com
 * released under GPL
 */
static char *rcsId = "$Id: msx.c,v 1.51 2009/02/18 14:17:36 folkert Exp folkert $";

#include <z80ex/z80ex.h>
#include <z80ex/z80ex_dasm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <ncurses.h>
#include <SDL/SDL.h>
#include <sndfile.h>

#include "error.h"
#include "gen.h"
#include "keyboard.h"
#include "video.h"
#include "disk.h"
#include "sound.h"
#include "memory.h"
#include "ports.h"

char verbose_mem = 0;
char verbose_mem_conf = 0;
char verbose_mem_mapper = 0;
char verbose_mem_megarom = 0;
char verbose_port = 0;
char verbose_cpu = 0;
char verbose_timing = 0;
char verbose_vdp = 0;
char verbose_sound = 0;
char verbose_disk = 1;
char verbose_rtc = 0;

char verbose_subsys = 0;

char verbose_mem_err = 1;
int verbose = 0;

char fullspeed = 0;
char vsync_redraw = 1;

int msx_hz = 3579545;
int REFRESH_RATE = 50;

int msx_level = 1;

char scroller = 1;

Z80EX_BYTE read_interrupt_vector(Z80EX_CONTEXT *cpu, void *user_data)
{
	if (verbose) fprintf(stderr, "read interrupt vector\n");
	return 0;
}

void show_cpu(int pc)
{
	int dummy;
	char buffer[128];
	int page, slot, sub_page, mem;
	unsigned char *p;
	split_addr(pc, &page, &slot, &sub_page, &mem);

	z80ex_dasm(buffer, sizeof(buffer), 0, &dummy, &dummy, read_mem_dis, pc, NULL);

	mvwprintw(stdscr, 1, 0, "PC: %04x, slot-select: %02x, sub-page: %02x", pc, ports[0xa8], sub_page_selection[slot]);
	mvwprintw(stdscr, 2, 0, "instruction: %s            ", buffer);

	fprintf(stderr, "%04x %s\n", pc, buffer);
}

void show_cpu_after(Z80EX_CONTEXT *z80)
{
	mvwprintw(stdscr, 3, 0, "after: AF: %04x, BC: %04x, DE: %04x, HL: %04x, PC: %04x, SP: %04x, IX: %04x, IY: %04x", z80ex_get_reg(z80, regAF), z80ex_get_reg(z80, regBC), z80ex_get_reg(z80, regDE), z80ex_get_reg(z80, regHL), z80ex_get_reg(z80, regPC), z80ex_get_reg(z80, regSP), z80ex_get_reg(z80, regIX), z80ex_get_reg(z80, regIY));
}

void show_vdp(void)
{
	int bg_map;

	if (video_mode < 4)
		bg_map = vdp_registers[2] << 10;
	else if (video_mode == 5 || video_mode == 6)
		bg_map = (vdp_registers[2] & 0x60) << 9;
	else
		bg_map = vdp_registers[2];

	mvwprintw(stdscr, 8, 0, "VDP 0: %02x, 1: %02x, mode: %d, pointer: %04x, bgmap: %04x, bgcolors: %04x, bgtiles: %04x",
			vdp_registers[0], vdp_registers[1], video_mode,
			vdp_rw_pointer + (video_mode >= 4 ? (vdp_registers[0x0e] & 3) << 8 : 0),
			bg_map, (vdp_registers[3] & (video_mode == 2 ? 128 : 255)) << 6, (vdp_registers[4] & (video_mode == 2 ? 4 : 255)) << 11);
}

void vsync(Z80EX_CONTEXT *z80, char *redraw, int *cycle_count, unsigned long long int *interrupts, unsigned long long int *int_cpu_not_ready, unsigned long long int *interrupts_ie0_not_set, char *flag_int_cpu_not_ready, double *last_vsync, unsigned long long int *total_cycles, double start, unsigned long long int *instructions)
{
	static double last_redraw = 0, last_scroll = 0;
	double last_vsync_duration = 1.0, now = get_ts(), run_time;
	int int_cycles;

	/* only check for SDL events and put them in the keyboard-queue
	 * which is checked in the read_ports() function
	 */
	check_keyboard();

	/* if VSYNC interrupts are allowed, try to service them. e.g. check if
	 * the processor is not currently processing an instruction. if it is,
	 * continue with the instruction and retry in the next iteration of the
	 * main loop
	 */
	if (vdp_registers[1] & 32) /* only interrupt when IE0 is set (bit 5) */
	{
		/* only reset 'timer' (well, we count cycles to know when the timer
		 * is triggered) when the interrupt was taken
		 */
		int_cycles = z80ex_int(z80);
		if (int_cycles)
		{
			if (vsync_redraw)
				*redraw = 1;
			*cycle_count = 0;
			(*interrupts)++;
			// wrong, must be status-reg vdp_registers[0] |= 128;
			// vdp_status_regs[0] |= 128; always set?
		}
		else
		{
			(*int_cpu_not_ready)++;
			*flag_int_cpu_not_ready = 1;
		}
		*total_cycles += int_cycles;

		last_vsync_duration = now - *last_vsync;
		*last_vsync = now;
	}
	else
	{
		*cycle_count = 0;
		if (vsync_redraw)
			*redraw = 1;

		(*interrupts_ie0_not_set)++;
	}

	vdp_status_regs[0] |= 128;

	/* some statistics */
	run_time = now - start;

	if (run_time)
		mvwprintw(stdscr, 5, 0, "run time: %f seconds, rel.speed: %6.2f%% / %6.2f%%, fps: %d, %.2f   ", run_time, (double)(*total_cycles * 100) / (double)(run_time * msx_hz), (100.0 / (double)REFRESH_RATE) / last_vsync_duration, fps, (double)*total_cycles / (double)*instructions);

	if ((now - last_redraw) >= 1.0)
	{
		*redraw = 1;
		last_redraw = now;

		mvwprintw(stdscr, 9, 20, "mapper: %02x %02x %02x %02x", mapper_bank_selection[0], mapper_bank_selection[1], mapper_bank_selection[2], mapper_bank_selection[3]);
	}

	if ((now - last_scroll) >= 0.25)
	{
		static char *scrolltext = "MSX-Emulator v" VERSION ", (C) 2008 by folkert@vanheusden.com Check www.vanheusden.com for other interesting stuff. UNIX sysop? Then see if www.multitail.org is of any help for you. This emulator was originally written in GFA basic for the Atari ST and was completely rewritten starting July 2008 for UNIX platforms altough it might compile and run on microsoft windows as well as I used the SDL library. The previous GFA basic version can be found (including all sourcecode) on my website.";
		static int pos = 0, dir = 1;
		int len;

		if (scroller)
		{
			len = strlen(scrolltext);
			SDL_WM_SetCaption(&scrolltext[pos], "");
			pos += dir;
			if (pos == len || pos == -1)
			{
				pos -= dir;
				dir = -dir;
			}
		}

		last_scroll = now;
	}

	/* in case the emulated version runs faster than the real one,
	 * sleep every vsync if difference > 10ms
	 */
	// FIXME: measure only the last timeslice (the time between the 2 vsyncs): the current
	// implementation won't work when we were slower than a real one for some time
	if (!fullspeed)
	{
		double should_have_taken_in_ms = ((double)*total_cycles / (double)msx_hz) * 1000.0;
		double diff = should_have_taken_in_ms - ((get_ts() - start) * 1000.0);
		if (diff > 10.0)
		{
		//	if (verbose) fprintf(stderr, "diff %f\n", diff);
			SDL_Delay(diff);
		}
	}
}

void loop(Z80EX_CONTEXT *z80)
{
	unsigned long long int total_cycles = 0, interrupts = 0, int_cpu_not_ready = 0, interrupts_ie0_not_set = 0, instructions = 0;
	int cycle_count = 0;
	double start = get_ts();
	double last_vsync = 0.0;

	mvwprintw(stdscr, 0, 0, "MSX-Emulator v" VERSION ", (C) 2008 by folkert@vanheusden.com");

	for(;;)
	{
		int cycles;
		char redraw = 0, flag_int_cpu_not_ready = 0;
		int pc = z80ex_get_reg(z80, regPC);

		if (cycle_count >= (msx_hz / REFRESH_RATE))
		{
			vsync(z80, &redraw, &cycle_count, &interrupts, &int_cpu_not_ready, &interrupts_ie0_not_set, &flag_int_cpu_not_ready, &last_vsync, &total_cycles, start, &instructions);
		}

		if (verbose_cpu || redraw)
			show_cpu(pc);

//		if (pc >= 0x4aa5 && pc < 0x5000)
//			show_cpu(pc);

		/* main z80 execution */
		cycles = z80ex_step(z80);
		cycle_count += cycles;
		total_cycles += cycles;
		instructions++;

		if (verbose_vdp || redraw)
			show_vdp();

		if (verbose_cpu || redraw)
			show_cpu_after(z80);

		if (verbose_timing || redraw || flag_int_cpu_not_ready) mvwprintw(stdscr, 4, 0, "cycles: %2d, total cycles: %lld, interrupts: %lld, cpu not ready: %lld, EI0=0: %lld ", cycles, total_cycles, interrupts, int_cpu_not_ready, interrupts_ie0_not_set);

		if ((redraw || verbose_subsys) && vdp_touched)
		{
			/* ncurses */
			redraw_ascii(stdscr, vdp_mem, vdp_registers);

			/* SDL */
			redraw_graphics(vdp_mem, vdp_registers);

			vdp_touched = 0;
		}
	}
}

void version(void)
{
	printf("MSX-Emulator v" VERSION ", (C) 2008-2010 by folkert@vanheusden.com\n\n");
}

void usage(void)
{
	version();

	printf("-r t:x:y:z:p  load rom z in slot x subslot y of type t, parameter p\n");
	printf("              t: bios (MSX bios/basic)\n");
	printf("                 rom  (generic ROM in page 1..3)\n");
	printf("                 exp  (MSX basic expanded)\n");
	printf("                 disk (DISK rom: requires disk-file for parameter p)\n");
	printf("                 megak4 (MegaROM of type Konami4 (Nemesis etc.)\n");
	printf("                 megak5 (MegaROM of type Konami5 (Nemesis 2 etc.)\n");
	printf("                 megam2 (MegaROM of type MSX-DOS2)\n");
	printf("                 ascii16kb (MegaROM of type ASCII 16KB)\n");
	printf("-m x:y  allocate a memory mapper in slot x subslot y\n");
	printf("-F      full speed (no delay loops)\n");
	printf("-f x    skip every xth frame while rendering display\n");
	printf("-G      only redraw screen once a second\n");
	printf("-L x    log vdp map-section output to file x\n");
	printf("-l x    MSX level (1=MSX1)\n");
	printf("-s      switch on sound\n");
	printf("-S x    store output in WAV-file x\n");
	printf("-v      increase verbosity (to stderr)\n");
	printf("-V      show version & exit\n");
	printf("-h      this help\n");
}

int main(int argc, char *argv[])
{
	int mp_slot = -1, mp_sub_slot = 0;
	int slot, page, sub_page;
	char *vdp_map_logfile = NULL;
	char **list = NULL;
	char megarom = 0;
	roms_t *roms = NULL;
	int n_roms = 0;
	Z80EX_CONTEXT *z80 = z80ex_create(read_mem_cb, NULL, write_mem_cb, NULL,
			read_port, NULL, write_port, NULL,
			read_interrupt_vector, NULL);
	int c;

	verbose_subsys = verbose_mem | verbose_mem_conf | verbose_port | verbose_cpu | verbose_timing | verbose_vdp;

	for(slot=0; slot<4; slot++)
	{
		for(page=0; page<4; page++)
		{
			for(sub_page=0; sub_page<4; sub_page++)
			{
				memory[slot][page][sub_page] = NULL;
				rom[slot][page][sub_page] = 0;
			}
		}
	}
	for(slot=0; slot<4; slot++)
		has_subpages[slot] = 0;

	while((c = getopt(argc, argv, "f:S:sr:m:FGL:l:QvVh")) != -1)
	{
		switch(c)
		{
			case 'Q':
				scroller = 0;
				break;

			case 'f':
				skip_frame = atoi(optarg);
				break;

			case 'r':
				split(optarg, &list);
				roms = (roms_t *)realloc(roms, sizeof(roms_t) * (n_roms + 1));
				memset(&roms[n_roms], 0x00, sizeof(roms_t));
				megarom = 0;
				if (strcmp(list[0], "bios") == 0)
					roms[n_roms].type = rom_bios;
				else if (strcmp(list[0], "rom") == 0)
					roms[n_roms].type = rom_generic;
				else if (strcmp(list[0], "exp") == 0)
					roms[n_roms].type = rom_msx2exp;
				else if (strcmp(list[0], "disk") == 0)
					roms[n_roms].type = rom_diskrom;
				else if (strcmp(list[0], "megak4") == 0)
				{
					roms[n_roms].type = rom_megarom_konami4;
					megarom = 1;
				}
				else if (strcmp(list[0], "megak5") == 0)
				{
					roms[n_roms].type = rom_megarom_konami5;
					megarom = 1;
				}
				else if (strcmp(list[0], "megam2") == 0)
				{
					roms[n_roms].type = rom_megarom_msxdos2;
					megarom = 1;
				}
				else if (strcmp(list[0], "ascii16kb") == 0)
				{
					roms[n_roms].type = rom_ascii_16kb;
					megarom = 1;
				}
				else
					error_exit("rom type %s not known", list[0]);
				roms[n_roms].slot = atoi(list[1]);
				roms[n_roms].subslot = atoi(list[2]);
				roms[n_roms].file = list[3];
				roms[n_roms].parameter = list[4];

				if (megarom)
				{
					megaroms = (megarom_t *)realloc(megaroms, (n_megaroms + 1)* sizeof(megarom_t));
					memset(&megaroms[n_megaroms], 0x00, sizeof(megarom_t));
					megaroms[n_megaroms].megarom_slot = roms[n_roms].slot;
					megaroms[n_megaroms].megarom_sub_slot = roms[n_roms].subslot;
					megaroms[n_megaroms].megarom_mapper_type = roms[n_roms].type;
					megaroms[n_megaroms].megarom_bank_selection[0] = 0;
					megaroms[n_megaroms].megarom_bank_selection[1] = 1;
					megaroms[n_megaroms].megarom_bank_selection[2] = 2;
					megaroms[n_megaroms].megarom_bank_selection[3] = 3;
					rom[roms[n_roms].slot][0][roms[n_roms].subslot] = rom[roms[n_roms].slot][1][roms[n_roms].subslot] = rom[roms[n_roms].slot][2][roms[n_roms].subslot] = rom[roms[n_roms].slot][3][roms[n_roms].subslot] = 3;
					n_megaroms++;
				}

				n_roms++;
				break;

			case 'l':
				msx_level = atoi(optarg);
				break;

			case 'm':
				split(optarg, &list);
				mp_slot = atoi(list[0]);
				mp_sub_slot = atoi(list[1]);

				mappers = (mapper_t *)realloc(mappers, sizeof(mapper_t) * (n_mappers + 1));

				mappers[n_mappers].mapper = (unsigned char *)malloc(16384 * 256);
				mappers[n_mappers].mapper_slot = mp_slot;
				mappers[n_mappers].mapper_sub_slot = mp_sub_slot;
				fprintf(stderr, "allocated 256 memory mapper pages at slot %d:%d\n", mappers[n_mappers].mapper_slot, mappers[n_mappers].mapper_sub_slot);

				has_subpages[mp_slot] |= mp_sub_slot > 0;
				rom[mp_slot][0][mp_sub_slot] = rom[mp_slot][1][mp_sub_slot] = rom[mp_slot][2][mp_sub_slot] = rom[mp_slot][3][mp_sub_slot] = 2;

				n_mappers++;
				break;

			case 'G':
				vsync_redraw = 0;
				break;

			case 'L':
				vdp_map_logfile = optarg;
				break;

			case 'F':
				fullspeed = 1;
				break;

			case 's':
				sound = 1;
				break;

			case 'S':
				soundfile = optarg;
				break;

			case 'v':
				verbose++;
				break;

			case 'h':
				usage();
				return 0;

			case 'V':
				version();
				return 0;

			default:
				usage();
				return 1;
		}
	}

	memset(ports, 0x00, sizeof(ports));
	memset(vdp_registers, 0x00, sizeof(vdp_registers));
	//	memset(rows_checked, 0x00, sizeof(rows_checked));

	if (n_mappers == 0)
		fprintf(stderr, "WARNING: no memory mappers selected\n");
	else
		fprintf(stderr, "Allocated %d mappers\n", n_mappers);

	init_memory(roms, n_roms);

	init_video(vdp_map_logfile);

	loop(z80);

	return 0;
}
