/* msx-emul is (C) 2008-2009 by folkert@vanheusden.com
 * released under GPL
 */
static char *rcsId = "$Id: ports.c,v 1.13 2009/02/18 14:18:40 folkert Exp folkert $";

#include <z80ex/z80ex.h>
#include <stdio.h>
#include <ncurses.h>
#include <SDL/SDL.h>
#include <sndfile.h>

#include "gen.h"
#include "msx.h"
#include "video.h"
#include "sound.h"
#include "keyboard.h"
#include "rtc.h"
#include "memory.h"
#include "hbi-55.h"

extern char verbose_port;
extern char verbose_mem_conf;

unsigned char ports[256];

Z80EX_BYTE read_port(Z80EX_CONTEXT *cpu, Z80EX_WORD port, void *user_data)
{
	unsigned char data = 0;

	port &= 0xff; /* on MSX, only the lower 8 bits are used */

	if (verbose_port || (verbose_mem_conf && port == 0xa8)) mvwprintw(stdscr, 7, 0, "%02x", port);

	switch(port)
	{
		case 0x90:			/* ULA5RA087 Centronic BUSY state (bit 1=1) */
			data = 0x00;		/* printer is always ready */
			break;

		case 0x98:
		case 0x99:
			data = read_vdp_port(port);
			break;

		case 0xa0:			/* PSG index (write only) */
			/* won't work */
			break;

		case 0xa1:			/* PSG data write */
			/* won't work */
			break;

		case 0xa2:			/* PSG data read(!) */
			if (psg_index < 0x0e)
				data = psg_regs[psg_index];
			else
				data = 0x3f;	/* because joystick is not finished yet FIXME */
			break;

		case 0xa4:			/* some turbo-r counter */
			data = ports[port]++;
			break;

		case 0xa7:			/* turbo-r pause button */
			data = 0x00; /* not pressed */
			break;

		case 0xa8:			/* slot select */
			data = ports[port];
			break;

		case 0xa9:				/* PPI port B: keyboard column inputs (RO) */
			data = read_keyboard(ports[0xaa] & 15);
			{ static int cnt = 0; mvwprintw(stdscr, 7, 0, "%02x", cnt++); refresh(); }
			break;

		case 0xaa:			/* ppi port c */
			data = ports[port];
			break;

		case 0xb0:			/* sony HBI-55 sram */
		case 0xb1:			/* sony HBI-55 */
		case 0xb2:			/* sony HBI-55 */
		case 0xb3:			/* sony HBI-55 */
			data = read_hbi_55(port);
			break;

		case 0xb4:			/* RTC index register (write only...) */
			data = rtc_index;
			break;

		case 0xb5:			/* rtc data register (read/write) */
			data = rtc_read(rtc_index);
			break;

		case 0xba:			/* FIXME */
		case 0xbb:
			break;

		case 0xe5:			/* turbo-r s1990 register */
			if (ports[port] == 0x05)
				return 0;	/* switch right: 0, left: 64 */
			else if (ports[port] == 0x06) /* mode, 96=z80, 64=r800, 0=r800/dram */
				return 96;
			data = -1;
			break;

		case 0xe6:			/* turbo-r s1990 counter */
		case 0xe7:
			{
				int counter = (((ports[0xe7] << 8) + ports[0xe6]) + 1) & 0xffff;
				ports[0xe7] = (counter >> 8);
				ports[0xe6] = counter & 255;

				if (port == 0xe7)
					data = ports[0xe7];
				else
					data = ports[0xe6];
			}
			break;

		case 0xf7:			/* audio/video control (r/w) */
			data = (ports[port] & 63) +
			       (((vdp_registers[9] >> 4) & 1) * 64) +
			       (((vdp_registers[9] >> 5) & 1) * 128);
			break;

		case 0xfc:
		case 0xfd:
		case 0xfe:
		case 0xff:
			data = mapper_bank_selection[port - 0xfc];
			break;

		default:
			data = ports[port];
			if (verbose) fprintf(stderr, "%f read unhandled port %02x (%02x)\n", get_ts(), port, data);
			break;
	}

	return data;
}

void write_port(Z80EX_CONTEXT *cpu, Z80EX_WORD port, Z80EX_BYTE value, void *user_data)
{
	port &= 0xff; /* on MSX, only the lower 8 bits are used */

	if (verbose_port || (verbose_mem_conf && port == 0xa8)) mvwprintw(stdscr, 7, 5, "| %02x %02x", port, value);

	ports[port] = value;

	switch(port)
	{
		case 0x01:				/* for debugging */
			if (value == 1)
				exit(0);
			break;

		case 0x91:				/* ULA5RA087 Centronic Printer Data */
			{
				FILE *fh = fopen("printer.txt", "a+");
				if (fh)
				{
					fprintf(fh, "%c", value);
					fclose(fh);
				}
			}
			break;

		case 0x98:	/* vram data */
		case 0x99:	/* register setup */
		case 0x9a:	/* palette register */
		case 0x9b:	/* register data */
			write_vdp_port(port, value);
			vdp_touched = 1;
			break;

		case 0xa0:			/* PSG index */
			psg_index = value & 15;
			break;

		case 0xa1:			/* PSG data write */
			psg_regs[psg_index] = value;
			if (psg_index < 0x0e && sound == 1)
				psg_sound(psg_regs, psg_index < 0x0e);
			break;

		case 0xa2:			/* PSG data read(!) */
			/* won't work */
			break;

		case 0xa7:			/* turbo-r pause/turbo leds */
			fprintf(stderr, "pause led %s\n", value & 1 ? "on" : "off");
			fprintf(stderr, "turbo led %s\n", value & 128 ? "on" : "off");
			break;

		case 0xa8:			/* slot select */
			break;

		case 0xaa:			/* PPI port C: keyboard row etc. */
			{ static int cnt = 0; mvwprintw(stdscr, 7, 8, "%02x", cnt++); }
			if (value & 128)	/* keyboard click bit */
				flash();
			break;

		case 0xab:			/* PPI control register */
			if (verbose) fprintf(stderr, "PPI control register %02x\n", value);

			if ((value & 128) == 0)
			{
				int bitnr = (value >> 1) & 7;

				if (value & 1)	/* 1=set */
					ports[0xaa] |= (1 << bitnr);
				else
					ports[0xaa] &= ~(1 << bitnr);
			}
			break;

		case 0xb0:			/* sony hbi-55 sram */
		case 0xb1:			/* sony hbi-55 sram */
		case 0xb2:			/* sony hbi-55 sram */
		case 0xb3:			/* sony hbi-55 sram */
			write_hbi_55(port, value);
			break;

		case 0xb4:			/* RTC index register (write only) */
			rtc_index = value;
			mvwprintw(stdscr, 9, 40, "idx %1x", rtc_index);
			break;

		case 0xb5:			/* rtc data register (read/write) */
			rtc_write(rtc_index, value);
			break;

		case 0xba:			/* FIXME */
		case 0xbb:
			break;

		case 0xe5:			/* Turbo-R s1990 register */
			break;

		case 0xe6:			/* s1990 reset counter */
			ports[0xe6] = ports[0xe7] = 0x00;
			break;

		case 0xfc:
		case 0xfd:
		case 0xfe:
		case 0xff:
			mapper_bank_selection[port - 0xfc] = value;
			break;

		default:
			if (verbose) fprintf(stderr, "write unhandled port %02x (%02x)\n", port, ports[port]);
			/* */
			break;
	}
}
