/* msx-emul is (C) 2008-2009 by folkert@vanheusden.com
 * released under GPL
 */
static char *rcsId = "$Id: rtc.c,v 1.5 2009/02/18 14:19:16 folkert Exp folkert $";

#include <time.h>
#include <ncurses.h>
#include <stdio.h>

#include "msx.h"

extern char verbose_rtc;

int rtc_index = 0;
unsigned char rtc_blocks[16][4];

unsigned char rtc_read(int index)
{
	int block = rtc_blocks[0x0d][0] & 3; /* block nr is in all blocks the same */

	if (block > 0 || index >= 0x0d)
	{
		if (verbose_rtc) fprintf(stderr, "RTC read block %d index %x\n", block, index);
		return (rtc_blocks[index][block] & 15) | 0xf0;
	}
	else
	{
		unsigned char data = -1;
		time_t now = time(NULL);
		struct tm *ptm = localtime(&now);

		if (verbose_rtc) fprintf(stderr, "RTC REAL read block %d index %x\n", block, index);

		switch(index)
		{
			case 0:		/* seconds low */
				data = ptm -> tm_sec & 15; break;
			case 1:		/* seconds high */
				data = ptm -> tm_sec >> 4; break;
			case 2:		/* minutes low */
				data = ptm -> tm_min & 15; break;
			case 3:		/* minutes high */
				data = ptm -> tm_min >> 4; break;
			case 4:		/* hour low */
				data = ptm -> tm_hour & 15; break;
			case 5:		/* hour high */
				data = ptm -> tm_hour >> 4; break;
			case 6:		/* day of week */
				data = ptm -> tm_wday; break;
			case 7:		/* day low */
				data = ptm -> tm_mday & 15; break;
			case 8:		/* day high */
				data = ptm -> tm_mday >> 4; break;
			case 9:		/* month low */
				data = ptm -> tm_mday & 15; break;
			case 0x0a:	/* month high */
				data = ptm -> tm_mday >> 4; break;
			case 0x0b:	/* year low */
				data = (ptm -> tm_year - 80) & 15; break;
			case 0x0c:	/* year high */
				data = (ptm -> tm_year - 80) >> 4; break;
		}

		return data | 0xf0;
	}

	if (verbose) fprintf(stderr, "rtc_read: unhandled case %d\n", index);

	return -1;
}

void rtc_write(int index, unsigned char data)
{
	int block = rtc_blocks[0x0d][0] & 3; /* block nr is in all blocks the same */

	if (block > 0 || index >= 0x0d)
	{
		if (index >= 0x0d)
			block = 0;

		rtc_blocks[index][block] = data;

		if (verbose_rtc) fprintf(stderr, "RTC write block %d index %x: %02x\n", block, index, data);

		if (index == 0x0f)	/* reset register */
		{
			int loop;

			for(loop=0; loop<16; loop++)
				rtc_blocks[loop][1] = 0;
		}
		else if (index == 0x0d) /* block */
		{
			mvwprintw(stdscr, 9, 40, "idx %1x blck: %d", index, block);
		}
	}
	else
	{
		if (verbose_rtc) fprintf(stderr, "Not setting time, index %d: %02x\n", index, data);
	}
}
