/* msx-emul is (C) 2008 by folkert@vanheusden.com
 * released under GPL
 */
static char *rcsId = "$Id: sound.c,v 1.5 2008/08/13 19:31:50 folkert Exp folkert $";

#include <SDL/SDL.h>
#include <math.h>
#include <sndfile.h>

#include "error.h"
#include "gen.h"
#include "msx.h"

char *soundfile = NULL;
SNDFILE *sndfile_out = NULL;
SF_INFO file_info;

unsigned char psg_regs[16];
unsigned char psg_index = 0;

unsigned char scc_regs[256];
char *scc_regs_s = (char *)scc_regs;

#define BASE_FREQ	(3579545 / 16)	/* clockfreq. psg */
#define BASE_FREQ_SCC	(3579545 / 32)	/* clockfreq. scc */
#define FREQ		22050		/* output samplerate */

short samples[FREQ * 2];
/* used by callback routine */
char initted = 0;
long int t = 0;
char sound = 0, scc_enabled = 0;

double pi = 3.14159265358979;
int nn_a, nn_b, nn_c;
double freq_a, freq_b, freq_c;
double mul_a, mul_b, mul_c;
double vol_a, vol_b, vol_c;
char use_a = 0, use_b = 0, use_c = 0;
char use_a_noise = 0, use_b_noise = 0, use_c_noise = 0;

char use_scc_1 = 0, use_scc_2 = 0, use_scc_3 = 0, use_scc_4 = 0, use_scc_5 = 0;
double vol_scc_1, vol_scc_2, vol_scc_3, vol_scc_4, vol_scc_5;
double mul_scc_1, mul_scc_2, mul_scc_3, mul_scc_4, mul_scc_5;

/* Prototype of our callback function */
void my_audio_callback(void *userdata, Uint8 *stream, int len)
{
	if (sound)
	{
		short *samples = (short *)stream;
		int nSamples = len / (2 * sizeof(short)), loop;

		// fprintf(stderr, "@my_audio_callback %f: %d (%d)|%d %d\n", get_ts(), nSamples, len, sound, scc_enabled);

		if ((psg_regs[0x08] & 15) == 0 && (psg_regs[0x09] & 15) == 0 && (psg_regs[0x0a] & 15) == 0 &&
				(!scc_enabled || ((scc_regs[0x8a] & 15) == 0 && (scc_regs[0x8b] & 15) == 0 && (scc_regs[0x8c] & 15) == 0 && (scc_regs[0x8d] & 15) == 0 && (scc_regs[0x8e] & 15) == 0)))
		{
			t = 0;
			// fprintf(stderr, " reset t\n");
		}

		for(loop=0; loop<nSamples; loop++)
		{
			double a = 0, b = 0, c = 0;
			double td = (double)t;

			if (use_a)
				a = (vol_a * sin(td * mul_a));
			if (use_b)
				b = (vol_b * sin(td * mul_b));
			if (use_c)
				c = (vol_c * sin(td * mul_c));

			if (scc_enabled)
			{
				double out[11], left = 0, right = 0;
				int oi = 0, ois, div;

				if (use_a_noise)
					out[oi++] = vol_a * (drand48() * 2.0 - 1.0);
				if (use_b_noise)
					out[oi++] = vol_b * (drand48() * 2.0 - 1.0);
				if (use_scc_1)
					out[oi++] = (double)scc_regs_s[0x00 + ((int)(td * mul_scc_1) & 0x1f)] * (vol_scc_1 / 128.0);
				if (use_scc_2)
					out[oi++] = (double)scc_regs_s[0x20 + ((int)(td * mul_scc_2) & 0x1f)] * (vol_scc_2 / 128.0);
				if (use_scc_3)
					out[oi++] = (double)scc_regs_s[0x40 + ((int)(td * mul_scc_3) & 0x1f)] * (vol_scc_3 / 128.0);
				if (use_scc_4)
					out[oi++] = (double)scc_regs_s[0x60 + ((int)(td * mul_scc_4) & 0x1f)] * (vol_scc_4 / 128.0);
				if (use_scc_5)
					out[oi++] = (double)scc_regs_s[0x60 + ((int)(td * mul_scc_5) & 0x1f)] * (vol_scc_5 / 128.0);
				out[oi++] = a;
				out[oi++] = b;
				out[oi++] = c;
				if (use_c_noise)
					out[oi++] = vol_c * (drand48() * 2.0 - 1.0);
				div = 0;
				for(ois=0; ois<(oi / 2); ois++) {
					left += out[ois];
					div++;
				}
				left /= (double)div;
				div = 0;
				for(;ois<oi; ois++) {
					right += out[ois];
					div++;
				}
				right /= (double)div;
				samples[(loop * 2) + 0] = (short)left;
				samples[(loop * 2) + 1] = (short)right;
			}
			else
			{
				if (use_a_noise)
				{
					a += vol_a * (drand48() * 2.0 - 1.0);
					a /= 2.0;
				}
				if (use_b_noise)
				{
					b += vol_b * (drand48() * 2.0 - 1.0);
					b /= 2.0;
				}
				if (use_c_noise)
				{
					c += vol_c * (drand48() * 2.0 - 1.0);
					c /= 2.0;
				}

				samples[(loop * 2) + 0] = (short)((a + b) / 2.0);
				samples[(loop * 2) + 1] = (short)((b + c) / 2.0);
			}

			t++;
		}

		if (sndfile_out != NULL && sf_write_short(sndfile_out, samples, nSamples * 2) != nSamples * 2)
			fprintf(stderr, "Failure writing %d samples to wav-file: %s\n", nSamples * 2, sf_strerror(sndfile_out));
	}
}

void init_audio(void)
{
	SDL_AudioSpec desired, obtained;

	desired.freq=FREQ;
	desired.format=AUDIO_S16SYS; /* AUDIO_S16LSB; */
	desired.channels=2;
	desired.samples=256; /*FIXME: (freq/100 / * tonen minstens 0.01s lang * /) en daar dan de onderste macht van 2 van */
	desired.callback=my_audio_callback;
	desired.userdata=NULL;

	/* Open the audio device */
	if (SDL_OpenAudio(&desired, &obtained) < 0)
		error_exit("Couldn't open audio: %s ", SDL_GetError());

	if (obtained.freq != FREQ || obtained.format != AUDIO_S16LSB || obtained.channels != 2)
		error_exit("requested audio type not available %d %d %d", obtained.freq, obtained.format, obtained.channels);

	SDL_PauseAudio(0);

	if (soundfile)
	{
		file_info.samplerate = FREQ;
		file_info.channels = 2;
		file_info.format = SF_FORMAT_WAV|SF_FORMAT_PCM_16;

		sndfile_out = sf_open(soundfile, SFM_WRITE, &file_info);
		if (!sndfile_out)
			fprintf(stderr, "Error opening wav-file %s\n", soundfile);
		else
			fprintf(stderr, "File %s created\n", soundfile);
	}
}

void psg_sound(unsigned char *regs, char audio_changed)
{
	if (!initted)
	{
		init_audio();
		initted = 1;
	}

	nn_a = ((psg_regs[1] & 15) << 8) + psg_regs[0];
	nn_b = ((psg_regs[3] & 15) << 8) + psg_regs[2];
	nn_c = ((psg_regs[5] & 15) << 8) + psg_regs[4];
	freq_a = (double)BASE_FREQ / (double)nn_a;
	freq_b = (double)BASE_FREQ / (double)nn_b;
	freq_c = (double)BASE_FREQ / (double)nn_c;
	mul_a = (2.0 * pi * freq_a) / (double)FREQ;
	mul_b = (2.0 * pi * freq_b) / (double)FREQ;
	mul_c = (2.0 * pi * freq_c) / (double)FREQ;
	vol_a = (psg_regs[0x08] & 16) ? 16384 : ((32767.0 / 15.0) * (double)(psg_regs[0x08] & 15));
	vol_b = (psg_regs[0x09] & 16) ? 16384 : ((32767.0 / 15.0) * (double)(psg_regs[0x09] & 15));
	vol_c = (psg_regs[0x0a] & 16) ? 16384 : ((32767.0 / 15.0) * (double)(psg_regs[0x0a] & 15));
	use_a = nn_a > 0 && (psg_regs[7] & 1) == 0 && vol_a > 0;
	use_b = nn_b > 0 && (psg_regs[7] & 2) == 0 && vol_b > 0;
	use_c = nn_c > 0 && (psg_regs[7] & 4) == 0 && vol_c > 0;
	use_a_noise = (psg_regs[7] &  8) == 0;
	use_b_noise = (psg_regs[7] & 16) == 0;
	use_c_noise = (psg_regs[7] & 32) == 0;

	// fprintf(stderr, "@psg_sound: vol %f %f %f\n", vol_a, vol_b, vol_c);
	// fprintf(stderr, "@psg_sound: freq %f %f %f\n", freq_a, freq_b, freq_c);
}

void scc_sound(void)
{
	int nn_1 = ((scc_regs[0x81] & 15) << 8) + scc_regs[0x80];
	int nn_2 = ((scc_regs[0x83] & 15) << 8) + scc_regs[0x82];
	int nn_3 = ((scc_regs[0x85] & 15) << 8) + scc_regs[0x84];
	int nn_4 = ((scc_regs[0x87] & 15) << 8) + scc_regs[0x86];
	int nn_5 = ((scc_regs[0x89] & 15) << 8) + scc_regs[0x88];
	double freq_1 = (double)BASE_FREQ_SCC / (double)nn_1;
	double freq_2 = (double)BASE_FREQ_SCC / (double)nn_2;
	double freq_3 = (double)BASE_FREQ_SCC / (double)nn_3;
	double freq_4 = (double)BASE_FREQ_SCC / (double)nn_4;
	double freq_5 = (double)BASE_FREQ_SCC / (double)nn_5;
	mul_scc_1 = (double)freq_1 / (double)FREQ;
	mul_scc_2 = (double)freq_2 / (double)FREQ;
	mul_scc_3 = (double)freq_3 / (double)FREQ;
	mul_scc_4 = (double)freq_4 / (double)FREQ;
	mul_scc_5 = (double)freq_5 / (double)FREQ;
	mul_scc_1 *= 32.0;
	mul_scc_2 *= 32.0;
	mul_scc_3 *= 32.0;
	mul_scc_4 *= 32.0;
	mul_scc_5 *= 32.0;
	vol_scc_1 = ((32767.0 / 15.0) * (double)(scc_regs[0x8a] & 15));
	vol_scc_2 = ((32767.0 / 15.0) * (double)(scc_regs[0x8b] & 15));
	vol_scc_3 = ((32767.0 / 15.0) * (double)(scc_regs[0x8c] & 15));
	vol_scc_4 = ((32767.0 / 15.0) * (double)(scc_regs[0x8d] & 15));
	vol_scc_5 = ((32767.0 / 15.0) * (double)(scc_regs[0x8e] & 15));
	use_scc_1 = scc_regs[0x8f] & 1 ? 1 : 0;
	use_scc_2 = scc_regs[0x8f] & 2 ? 1 : 0;
	use_scc_3 = scc_regs[0x8f] & 4 ? 1 : 0;
	use_scc_4 = scc_regs[0x8f] & 8 ? 1 : 0;
	use_scc_5 = scc_regs[0x8f] & 16 ? 1 : 0;

	if (!initted)
	{
		init_audio();
		initted = 1;
	}
}
