/* msx-emul is (C) 2008-2009 by folkert@vanheusden.com
 * released under GPL
 */
static char *rcsId = "$Id: video.c,v 1.29 2009/02/19 14:33:55 folkert Exp folkert $";

#include <z80ex/z80ex.h>
#include <ncurses.h>
#include <SDL/SDL.h>

#include "error.h"
#include "gen.h"
#include "msx.h"
#include "ports.h"

int skip_frame = 0;
unsigned char *vdp_mem;
int vdp_rw_pointer = 0;
char vdp_addr_state = 0;
unsigned char vdp_addr_b1 = 0;
unsigned char vdp_registers[256]; /* altough msx has 8 and msx2 has 2f of them */
unsigned char vdp_status_regs[256]; /* msx1: 1, msx2: 9 */
int video_mode = 0;
unsigned char vdp_readahead = 0x00;
int fps = 0;
int rg[8], b[4];
int sourcex = 0, sourcey = 0, destinationx = 0, destinationy = 0, numberx = 0, numbery = 0;
int start_destinationx, start_destinationy;
char vdp_cmd = -1;
int pixelsleft = 0, pixeloffset = 0;
int scr_width = -1, ppb = 0; /* ppb = pixels per byte */
char vdp_touched = 0;
char highspeed = 0;
char kb64 = 0;
SDL_Surface *screen = NULL;
SDL_VideoInfo *svi;
int sat;
int sc;
int scp;

FILE *map_log_fh;

/* TMS9918 palette */
unsigned char rgb[16][3] = {
	 { 0, 0, 0 },
	 { 0, 0, 0 },
	 { 33, 200, 66 },
	 { 94, 220, 120 },
	 { 84, 85, 237 },
	 { 125, 118, 252 },
	 { 212, 82, 77 },
	 { 66, 235, 245 },
	 { 252, 85, 84 },
	 { 255, 121, 120 },
	 { 212, 193, 84 },
	 { 231, 206, 128 },
	 { 33, 176, 59 },
	 { 201, 91, 186 },
	 { 204, 204, 204 },
	 { 255, 255, 255 }
	 	};

unsigned char palette[256][3], pal_sel = 0, pal_byte_0 = 0;

unsigned char get_vdp_register(int register_index)
{
	vdp_reg_name(register_index);

	return vdp_registers[register_index];
}

void setPixelUint(int x, int y, Uint32 pixel) 
{ 
	int bpp = screen->format->BytesPerPixel; 

	Uint8 *p = (Uint8 *)screen->pixels + y * screen->pitch + x * bpp; 

	if (bpp == 3)
	{
		p[1] = (pixel >> 8);
		if(SDL_BYTEORDER == SDL_BIG_ENDIAN) { 
			p[0] = (pixel >> 16);
			p[2] = pixel;
		}
		else { 
			p[0] = pixel;
			p[2] = pixel >> 16;
		} 
	}
	else
	{
		*(Uint32 *)p = pixel; 
	} 
} 

void setPixel(int x, int y, int r, int g, int b)
{
	Uint32 color = SDL_MapRGB(screen->format, r, g, b);

	setPixelUint(x, y, color);
}

void lockScreen()
{
	/* Set the surface pixels and refresh! */
	if (SDL_MUSTLOCK(screen))
	{
		if (SDL_LockSurface(screen) < 0)
			error_exit("Couldn't lock the display surface: %s", SDL_GetError());
	}
}

void unlockScreen()
{
	if (SDL_MUSTLOCK(screen))
		SDL_UnlockSurface(screen);
}

void updateScreen()
{
	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

unsigned char read_vdp(unsigned char *vdp_mem, int pointer)
{
	if (verbose)
	{
		if ((pointer < 0 || pointer > 16383) && (msx_level == 1 || video_mode < 4))
			fprintf(stderr, "read: vdp mem overflow %d\n", pointer);
	}

	return vdp_mem[pointer];
}

void write_vdp(unsigned char *vdp_mem, int pointer, unsigned char value)
{
	if (verbose)
	{
		if ((pointer < 0 || pointer > 16383) && (msx_level == 1 || video_mode < 4))
			fprintf(stderr, "write: vdp mem overflow %d\n", pointer);
	}

	if (map_log_fh)
	{
		int bg_map      = (get_vdp_register(2) & 15) << 10;
		int bg_map_size = (video_mode == 1 ? 40 : 32) * 24;

		if (pointer >= bg_map && pointer < (bg_map + bg_map_size))
			fprintf(map_log_fh, "%c", value);
	}

	if (!kb64 || pointer < 65536)
		vdp_mem[pointer] = value;
}

void update_video_mode(void)
{
	int prev_video_mode = video_mode;

	// M1 M2 M3 M4 M5  Screen format
	// 1  0  0  0  0   Text       40x24             (BASIC SCREEN 0)
	// 0  0  0  0  0   Half text  32x24             (BASIC SCREEN 1)
	// 0  0  1  0  0   Hi resolution 256x192        (BASIC SCREEN 2)
	// 0  1  0  0  0   Multicolour  4x4pix blocks   (BASIC SCREEN 3)
	// ----Below MSX2 only----
	// 0  0  0  1  0   Screen2 with 8 Sprites/Line  (BASIC SCREEN 4)
	// 0  0  1  1  0   256*212, 16  colours/pixel   (BASIC SCREEN 5) == graphics mode 4 in v9938 msx-video techinal data book by ascii
	// 0  0  0  0  1   512*212, 4   colours/pixel   (BASIC SCREEN 6)
	// 0  0  1  0  1   512*212, 16  colours/pixel   (BASIC SCREEN 7)
	// 0  0  1  1  1   256*212, 256 colours/pixel   (BASIC SCREEN 8)
	// 1  0  0  1  0   Text 80x24                   (BASIC SCREEN 0, WIDTH 80)
	char m1 = (get_vdp_register(1) >> 4) & 1;
	char m2 = (get_vdp_register(1) >> 3) & 1;
	char m3 = (get_vdp_register(0) >> 1) & 1;
	char m4 = (get_vdp_register(0) >> 2) & 1;
	char m5 = (get_vdp_register(0) >> 3) & 1;

	if (m1 == 1 && m2 == 0 && m3 == 0 && m4 == 0 && m5 == 0)
		video_mode = 1;
	else if (m1 == 0 && m2 == 0 && m3 == 0 && m4 == 0 && m5 == 0)
		video_mode = 0;
	else if (m1 == 0 && m2 == 0 && m3 == 1 && m4 == 0 && m5 == 0) {
		scr_width = 256;
		video_mode = 2;
	}
	else if (m1 == 0 && m2 == 1 && m3 == 0 && m4 == 0 && m5 == 0)
		video_mode = 3;
	else if (m1 == 0 && m2 == 0 && m3 == 0 && m4 == 1 && m5 == 0 && msx_level > 1) {
		scr_width = 256;
		video_mode = 4;
		ppb = 8;
	}
	else if (m1 == 0 && m2 == 0 && m3 == 1 && m4 == 1 && m5 == 0 && msx_level > 1) {
		scr_width = 256;
		video_mode = 5;
		ppb = 2;
	}
	else if (m1 == 0 && m2 == 0 && m3 == 0 && m4 == 0 && m5 == 1 && msx_level > 1) {
		scr_width = 512;
		video_mode = 6;
		ppb = 4;
	}
	else if (m1 == 0 && m2 == 0 && m3 == 1 && m4 == 0 && m5 == 1 && msx_level > 1) {
		scr_width = 512;
		video_mode = 7;
		ppb = 2;
	}
	else if (m1 == 0 && m2 == 0 && m3 == 1 && m4 == 1 && m5 == 1 && msx_level > 1) {
		scr_width = 256;
		video_mode = 8;
		ppb = 1;
	}
	else if (m1 == 1 && m2 == 0 && m3 == 0 && m4 == 1 && m5 == 0 && msx_level > 1)
	{
		scr_width  = 640;
		video_mode = 80;
	}

	if (video_mode != prev_video_mode)
	{
		fprintf(stderr, "vdp switched from video mode %d to %d\n", prev_video_mode, video_mode);
		resize_window(scr_width, 212);
	}
}

int get_pixel(int video_mode, int x, int y, char highspeed)
{
	if (video_mode == 5)
	{
		int offset = (y * (256/2)) + (x / 2);
		int hnibble = x & 1;
		int shift = (1 - hnibble) * 4;
		unsigned char mask = 15;

		if (highspeed)
			return vdp_mem[offset];

		return (vdp_mem[offset] >> shift) & mask;
	}
	else if (video_mode == 6)
	{
		int offset = (y * (512/4)) + (x / 4);
		int hnibble = x & 3;
		int shift = (3 - hnibble) * 2;
		unsigned char mask = 3;

		if (highspeed)
			return vdp_mem[offset];

		return (vdp_mem[offset] >> shift) & mask;
	}
	else if (video_mode == 7)
	{
		int offset = (y * (512/2)) + (x / 2);
		int hnibble = x & 1;
		int shift = (1 - hnibble) * 4;
		unsigned char mask = 15;

		if (highspeed)
			return vdp_mem[offset];

		return (vdp_mem[offset] >> shift) & mask;
	}
	else if (video_mode == 8)
	{
		int offset = (y * 256) + x;

		if (highspeed)
			return vdp_mem[offset];

		return vdp_mem[offset];
	}
	// FIXME: other video-modes

	return -1;
}

void plot(int video_mode, int x, int y, int color)
{
	if (video_mode == 5)
	{
		int offset = (y * (256/2)) + (x / 2);
		if (highspeed)
			vdp_mem[offset] = color;
		else
		{
			int hnibble = x & 1;
			int shift = (1 - hnibble) * 4;
			unsigned char mask = ~(15 << shift);

			vdp_mem[offset] &= mask;
			vdp_mem[offset] |= (color & 15) << shift;
		}
	}
	else if (video_mode == 6)
	{
		int offset = (y * (512/4)) + (x / 4);
		int hnibble = x & 3;
		int shift = (3 - hnibble) * 2;
		unsigned char mask = ~(3 << shift);

		vdp_mem[offset] &= mask;
		vdp_mem[offset] |= (color & 3) << shift;
	}
	else if (video_mode == 7)
	{
		int offset = (y * (512/2)) + (x / 2);
		if (highspeed)
			vdp_mem[offset] = color;
		else
		{
			int hnibble = x & 1;
			int shift = (1 - hnibble) * 4;
			unsigned char mask = ~(15 << shift);

			vdp_mem[offset] &= mask;
			vdp_mem[offset] |= (color & 15) << shift;
		}
	}
	else if (video_mode == 8)
	{
		int offset = (y * 256) + x;

		vdp_mem[offset] = color;
	}

	// FIXME: other video-modes
}

void put_vdp_2c(unsigned char value)
{
	if ((vdp_cmd == 0x05 || vdp_cmd == 0x0b) && pixelsleft > 0) /* logical put pixels, cpu -> vram */
	{
		int y = pixeloffset / numberx;
		int x = pixeloffset - (y * numberx);

		plot(video_mode, destinationx, destinationy, value);
		pixeloffset++;
		pixelsleft--;

		destinationx++;
		if (destinationx == (start_destinationx + numberx))
		{
			destinationx = start_destinationx;
			destinationy++;
		}
	}
	else if (vdp_cmd == 0x0f && pixelsleft > 0)
	{
		int np = 0;
		int dy = pixeloffset / numberx;
		int y = destinationy + dy;
		int x = destinationx + (pixeloffset - (dy * numberx));
		int offset = 0;

		if (video_mode == 5)
			offset = (y * (256/2)) + (x / 2);
		else if (video_mode == 6)
			offset = (y * (512/4)) + (x / 4);
		else if (video_mode == 7)
			offset = (y * (512/2)) + (x / 2);
		else if (video_mode == 8)
			offset = (y * 256) + x;

		vdp_mem[offset] = value;

		if (video_mode == 5)
			np = 2;
		else if (video_mode == 6)
			np = 4;
		else if (video_mode == 7)
			np = 2;
		else if (video_mode == 8)
			np = 1;
		// FIXME other videomodes

		pixeloffset += np;
		pixelsleft -= np;
	}
	else
		fprintf(stderr, "put_vdp_2c unsupported command %02x value %02x\n", vdp_cmd, value);

}

void draw_line(int video_mode, int destinationx, int destinationy, int numberx, int numbery, int color)
{
	double error = 0.0;
	double deltaerr = (double)numbery / (double)numberx;
	int y = destinationy, x;
	for(x=destinationx; x<(destinationx+numberx); x++)
	{
		plot(video_mode, x, y, color);
		error += deltaerr;
		if (abs(error) >= 0.5)
		{
			y++;
			error -= .0;
		}
	}
}

void fill_rect(int video_mode, int destinationx, int destinationy, int numberx, int numbery, int color)
{
	int x, y;

	for(y=destinationy; y<(destinationy + numbery); y++)
	{
		for(x=destinationx; x<(destinationx + numberx); x++)
		{
			plot(video_mode, x, y, color);
		}
	}
}

void copy_rect(int video_mode, int sourcex, int sourcey, int destinationx, int destinationy, int numberx, int numbery, char highspeed)
{
	int x, y;

	if (highspeed)
	{
		int plus = 1;

		if (video_mode == 5)
		{
			numberx |= 1;
			destinationx &= ~1;
			plus = 2;
		}
		else if (video_mode == 6)
		{
			numberx |= 3;
			destinationx &= ~3;
			plus = 4;
		}
		else if (video_mode == 7)
		{
			numberx |= 1;
			destinationx &= ~1;
			plus = 2;
		}
		else if (video_mode == 8)
		{
			// numberx |= 0;
			// destinationx &= ~0;
			plus = 1;
		}

		for(y=0; y<numbery; y++)
		{
			for(x=0; x<numberx; x+=plus)
			{
				plot(video_mode, destinationx + x, destinationy + y, get_pixel(video_mode, sourcex + x, sourcey + y, 1));
			}
		}
	}
	else
	{
		for(y=0; y<numbery; y++)
		{
			for(x=0; x<numberx; x++)
			{
				plot(video_mode, destinationx + x, destinationy + y, get_pixel(video_mode, sourcex + x, sourcey + y, 0));
			}
		}
	}
}

void vdp_reg_name(int register_index)
{
	if (register_index <= 0x2e)
	{
		char *names[] = {
			"mode register 0",	/* 00 */
			"mode register 1",	/* 01 */
			"bg map ",		/* 02 */
			"bg colors",		/* 03 */
			"bg tiles",		/* 04 */
			"obj attr",		/* 05 */
			"obj tiles",	 	/* 06 */
			"colour reg",		/* 07 */
			"mode reg 2",		/* 08 */
			"mode reg 3",		/* 09 */
			"colour high",		/* 0a */
			"sprite attr high",	/* 0b */
			"inverse color",	/* 0c */
			"blink period",		/* 0d */
			"vram access",		/* 0e */
			"status reg index",	/* 0f */
			"palette index",	/* 10 */
			"register pointer",	/* 11 */
			"display adjust",	/* 12 */
			"interrupt line",	/* 13 */
			"color burst 1",	/* 14 */
			"color burst 2",	/* 15 */
			"color burst 3",	/* 16 */
			"display offset",	/* 17 */
			"not used",		/* 18 */
			"horiz scroll",		/* 19 */
			"horiz scroll high",	/* 1a */
			"horiz scroll low",	/* 1b */
			"not used",		/* 1c */
			"not used",		/* 1d */
			"not used",		/* 1e */
			"not used",		/* 1f */
			"Source X Low byte",	/* 20 */
			"Source X High byte",	/* 21 */
			"Source Y Low byte",	/* 22 */
			"Source Y High byte",	/* 23 */
			"Destination X Low byte",	/* 24 */
			"Destination X High byte",	/* 25 */
			"Destination Y Low byte",	/* 26 */
			"Destination Y High byte",	/* 27 */
			"Number of X dots low byte",	/* 28 */
			"Number of X dots high byte",	/* 29 */
			"Number of Y dots low byte",	/* 2a */
			"Number of Y dots high byte",	/* 2b */
			"Data",			/* 2c */
			"argument register",	/* 2d */
			"command register"	/* 2e */
			};

		fprintf(stderr, "%s ", names[register_index]);
	}
}

void set_vdp_register(int register_index, unsigned char data)
{
	vdp_registers[register_index] = data;

	vdp_reg_name(register_index);

	switch(register_index)
	{
		case 9:
			fprintf(stderr, "%d\n", data);
			break;

		case 0x0e:			/* bank sel */
			fprintf(stderr, "vdp bank sel %02x\n", data);
			break;

		case 0x10:			/* palette index */
			pal_sel = 0;	/* reset 'pointer' to first byte (of two) */
			break;

		case 0x20:
			sourcex &= ~255;
			sourcex |= data;
			break;

		case 0x21:
			sourcex &= 255;
			sourcex |= (data & 1) << 8;
			break;

		case 0x22:
			sourcey &= ~255;
			sourcey |= data;
			break;

		case 0x23:
			sourcey &= 255;
			sourcey |= (data & 3) << 8;
			break;

		case 0x24:
			destinationx &= ~255;
			destinationx |= data;
			break;

		case 0x25:
			destinationx &= 255;
			destinationx |= (data & 1) << 8;
			break;

		case 0x26:
			destinationy &= ~255;
			destinationy |= data;
			break;

		case 0x27:
			destinationy &= 255;
			destinationy |= (data & 3) << 8;
			break;

		case 0x28:
			numberx &= ~255;
			numberx |= data;
			break;

		case 0x29:
			numberx &= 255;
			numberx |= (data & 3) << 8;
			break;

		case 0x2a:
			numbery &= ~255;
			numbery |= data;
			break;

		case 0x2b:
			numbery &= 255;
			numbery |= (data & 3) << 8;
			break;

		case 0x2c:			/* data register */
			fprintf(stderr, "vdp data register: %02x\n", data);
			break;

		case 0x2e:			/* command register */
			vdp_cmd = data >> 4;
			pixelsleft = numberx * numbery;
			pixeloffset = 0;

			fprintf(stderr, "@vdp[%d] command arg: %x, transp: %d, cmd: %02x |", video_mode, data & 7, (data >> 3) & 1, vdp_cmd);
			fprintf(stderr, "  source %d,%d - dest %d,%d | num %d,%d\n", sourcex, sourcey, destinationx, destinationy, numberx, numbery);

			start_destinationx = destinationx;
			start_destinationy = destinationy;

			highspeed = 0;
			switch(vdp_cmd)
			{
				case 0x05:
					put_vdp_2c(get_vdp_register(0x2c));
					break;

				case 0x07:
					draw_line(video_mode, destinationx, destinationy, numberx, numbery, get_vdp_register(0x2c));
					break;

				case 0x09:
					copy_rect(video_mode, sourcex, sourcey, destinationx, destinationy, numberx, numbery, 0);
					break;

				case 0x0b:	// CPU -> VRAM
					// there's alreay 1 pixel in the buffer, plot it
					put_vdp_2c(get_vdp_register(0x2c));
					// from now on untill all pixels have been plotted, pixels will be written to 0x2c
					highspeed = 0;
					break;

				case 0x0c:
					highspeed = 1;
					fill_rect(video_mode, destinationx, destinationy, numberx, numbery, get_vdp_register(0x2c));
					break;

				case 0x0d:
					highspeed = 1;
					copy_rect(video_mode, sourcex, sourcey, destinationx, destinationy, numberx, numbery, 1);
					break;

				case 0x0f:
					highspeed = 1;
					plot(video_mode, destinationx, destinationy, get_vdp_register(0x2c));
					break;

				default:
					fprintf(stderr, "unhandled vdp command\n");
					break;
			}

			break;
	}

	if (register_index == 0 || register_index == 1)
		update_video_mode();
}

void write_vdp_port(Z80EX_WORD port, Z80EX_BYTE value)
{
	switch(port)
	{
		case 0x98:					/* VDP data */
			if (msx_level > 1 && video_mode >= 4)
			{
				fprintf(stderr, "vdp data write %05x %02x\n", ((get_vdp_register(0x0e) & 7) << 14) + vdp_rw_pointer, value);
				write_vdp(vdp_mem, ((get_vdp_register(0x0e) & 7) << 14) + vdp_rw_pointer, value);

				vdp_rw_pointer++;
				if (vdp_rw_pointer == 16384)
				{
					vdp_rw_pointer = 0;

					vdp_registers[0x0e]++;
					vdp_registers[0x0e] &= 7; /* only lower 3 bits are used */
					fprintf(stderr, "vdp bank move %02x\n", vdp_registers[0x0e]);
				}
			}
			else
			{
				fprintf(stderr, "vdp data write %04x %02x\n", vdp_rw_pointer, value);
				write_vdp(vdp_mem, vdp_rw_pointer++, value);
				vdp_rw_pointer &= 16383;
			}

			vdp_readahead = value;
			vdp_addr_state = 0;
			break;

		case 0x99:					/* VDP address setup/register write */
			if (vdp_addr_state == 0)
			{
				vdp_addr_b1 = value;
			}
			else
			{
				if ((value & 128) == 128)	/* bit 7 == 1: register setup */
				{
					if (msx_level > 1)
						value &= 63; /* max 0x2e */
					else
						value &= 7;

					fprintf(stderr, "VDP write reg %02x val %02x\n", value, vdp_addr_b1);

					set_vdp_register(value, vdp_addr_b1);
				}
				else			/* memory */
				{
					vdp_rw_pointer = ((value & 63) << 8) + vdp_addr_b1;
					fprintf(stderr, "set vdp pointer to %04x\n", vdp_rw_pointer);

					if ((value & 64) == 0)
					{
						if (msx_level > 1 && video_mode >= 4)
							vdp_readahead = read_vdp(vdp_mem, ((get_vdp_register(0x0e) & 7) << 14) + vdp_rw_pointer);
						else
							vdp_readahead = read_vdp(vdp_mem, vdp_rw_pointer);
						vdp_rw_pointer++;
						fprintf(stderr, "VDP address setup, increase pointer to %04x, read data froma %04x\n", vdp_rw_pointer, vdp_rw_pointer-1);

						if (vdp_rw_pointer == 16384)
						{
							if (msx_level > 1 && video_mode >= 4)
							{
								vdp_registers[0x0e]++;
								vdp_registers[0x0e] &= 7; /* only lower 3 bits are used */
							}

							vdp_rw_pointer = 0;
						}
					}
				}
			}
			vdp_addr_state = !vdp_addr_state;
			break;

		case 0x9a:			/* set palette */
			if (pal_sel == 0)
			{
				pal_byte_0 = value;
			}
			else
			{
				int entry = get_vdp_register(0x10) & 15;

				palette[entry][2] = rg[pal_byte_0 & 7];		/* blue  */
				palette[entry][0] = rg[(pal_byte_0 >> 4) & 7];	/* red   */
				palette[entry][1] = rg[value & 7];		/* green */

				fprintf(stderr, "vdp set palette %d to %d %d %d\n",
						entry, palette[entry][0], palette[entry][1], palette[entry][2]);

				vdp_registers[0x10]++;
				vdp_registers[0x10] &= 15;
			}

			pal_sel = 1 - pal_sel;
			break;


		case 0x9b:
			{
				int register_index = get_vdp_register(0x11) & 63;

				fprintf(stderr, "0x9b: set VDP reg %02x to value %02x\n", register_index, value);

				if ((register_index) != 0x11)
					set_vdp_register(register_index, value);

				if ((get_vdp_register(0x11) & 128) == 0)
				{
					vdp_registers[0x11]++;
					vdp_registers[0x11] &= 63;
				}

				if (register_index == 0x2c)
					put_vdp_2c(value);
			}
			break;
	}
}

unsigned char read_vdp_port(int port)
{
	unsigned char data = -1;

	switch(port)
	{
		case 0x98:
			data = vdp_readahead;

			if (msx_level > 1 && video_mode >= 4)
			{
				fprintf(stderr, "vdp data read %05x %02x\n", ((get_vdp_register(0x0e) & 7) << 14) + vdp_rw_pointer, data);

				vdp_readahead = read_vdp(vdp_mem, ((get_vdp_register(0x0e) & 7) << 14) + vdp_rw_pointer);

				vdp_rw_pointer++;
				if (vdp_rw_pointer == 16384)
				{
					vdp_rw_pointer = 0;

					vdp_registers[0x0e]++;
					vdp_registers[0x0e] &= 7; /* only lower 3 bits are used */
					fprintf(stderr, "vdp bank move %02x\n", vdp_registers[0x0e]);
				}
			}
			else
			{
				fprintf(stderr, "vdp data read %05x %02x\n", vdp_rw_pointer, data);

				vdp_readahead = read_vdp(vdp_mem, vdp_rw_pointer++);
				vdp_rw_pointer &= 16383;
			}

			vdp_addr_state = 0;
			break;

		case 0x99:		/* vdp status register (read only) */
			data = vdp_status_regs[get_vdp_register(15)];
			if (vdp_registers[15] == 0)
				vdp_status_regs[vdp_registers[15]] &= 127;	/* read clears bit 7 (interrupt) */
			else if (vdp_registers[15] == 2)
				vdp_status_regs[2] = (vdp_status_regs[2] & 0x1e) | ((~(vdp_status_regs[2] & 0xe1)) & 0xe1);
			vdp_addr_state = 0;
			fprintf(stderr, "read vdp status reg %d: %02x\n", vdp_registers[15], data);
			break;
	}

	return data;
}

void redraw_ascii(WINDOW *win, unsigned char *vdp_mem, unsigned char *vdp_registers)
{
	int x, y;
	int cols;
	int bg_map = (video_mode == 80) ? (vdp_registers[2] & 0x7c) << 10 : (vdp_registers[2] & 15) << 10;

	if (video_mode == 1)
		cols = 40;
	else if (video_mode == 80)
		cols = 80;
	else
		cols = 32;

	for(y=0; y<24; y++)
	{
		for(x=0; x<cols; x++)
		{
			unsigned char obj = read_vdp(vdp_mem, bg_map + (y * cols) + x);
			mvwaddch(win, y + 11, x, obj?obj:' ');
		}
	}

	refresh();
}

void redraw_screen0(int cols)
{
	int scr_x, scr_y;
	int bg_map = (video_mode == 80) ? (vdp_registers[2] & 0x7c) << 10 : (vdp_registers[2] & 15) << 10;
	int bg_tiles = (vdp_registers[4] & 7) << 11;
	unsigned char fg_r, fg_g, fg_b;
	unsigned char bg_r, bg_g, bg_b;
	Uint32 colorbg, colorfg;

	bg_r = rgb[vdp_registers[7] & 15][0];
	bg_g = rgb[vdp_registers[7] & 15][1];
	bg_b = rgb[vdp_registers[7] & 15][2];
	fg_r = rgb[vdp_registers[7] >> 4][0];
	fg_g = rgb[vdp_registers[7] >> 4][1];
	fg_b = rgb[vdp_registers[7] >> 4][2];
	// fprintf(stderr, "%d %d %d | %d %d %d\n", bg_r, bg_g, bg_b, fg_r, fg_g, fg_b);
	colorbg = SDL_MapRGB(screen->format, bg_r, bg_g, bg_b);
	colorfg = SDL_MapRGB(screen->format, fg_r, fg_g, fg_b);

	for(scr_y=0; scr_y<24; scr_y++)
	{
		for(scr_x=0; scr_x<cols; scr_x++)
		{
			int map_index = bg_map + (scr_y * cols) + scr_x;
			int sel_char = read_vdp(vdp_mem, map_index);
			int tile = bg_tiles + (sel_char * 8);
			int char_x, char_y;

			for(char_y=0; char_y<8; char_y++)
			{
				int p = tile + char_y;
				unsigned char pixels = read_vdp(vdp_mem, p);

				for(char_x=0; char_x<8; char_x++)
				{
					setPixelUint(scr_x * 8 + char_x, scr_y * 8 + char_y, pixels & 128 ? colorfg : colorbg);

					pixels <<= 1;
				}
			}
		}
	}
}

void redraw_screen1()
{
	int bg_map    = (vdp_registers[2] &  15) << 10;
	int bg_colors = (vdp_registers[3] & 128) <<  6; /* in scr 2 only bit 7 is used, so offset is 0x2000 or 0x0 */
	int bg_tiles  = (vdp_registers[4] &   4) << 11; /* in scr 2 only bit 2 */
	int map_index, loop;
	Uint32 colors[16];

	// FIXME: kleur index 0: vdp reg 7 bg
	for(loop=0; loop<16; loop++)
		colors[loop] = SDL_MapRGB(screen->format, rgb[loop][0], rgb[loop][1], rgb[loop][2]);

	for(map_index=0; map_index<(32*24); map_index++)
	{
		int cur_char_nr = read_vdp(vdp_mem, bg_map + map_index);
		int cur_tiles   = bg_tiles + (cur_char_nr * 8);
		/* offset in SDL window */
		int scr_x       = (map_index & 31) * 8;
		int scr_y       = ((map_index >> 5) & 31) * 8;
		int x, y;
		unsigned char current_color = read_vdp(vdp_mem, bg_colors + cur_char_nr / 8);
		Uint32 fg       = colors[current_color >> 4];
		Uint32 bg       = colors[current_color & 15];

		for(y=8; y; y--, scr_y++)
		{
			unsigned char current_tile = read_vdp(vdp_mem, cur_tiles++);
			int end_x = scr_x + 8;

			for(x=scr_x; x<end_x; x++, current_tile <<= 1)
				setPixelUint(x, scr_y, current_tile & 128 ? fg : bg);
		}
	}
}

void draw_sprite_part(int off_x, int off_y, int pattern_offset, Uint32 color, int nr)
{
	int x, y, ynr =0;
	int width = screen -> w, height = screen -> h;
	int sc = ((vdp_registers[5] & ~3) << 7) + ((vdp_registers[11] & 3) << 15) - 0x200;

	for(y=off_y; y<(off_y + 8); y++)
	{
		int cur_pattern = read_vdp(vdp_mem, pattern_offset++);
		unsigned char col = read_vdp(vdp_mem, sc + nr * 16 + ynr);
		unsigned char i = col & 8 ? 128 : 0;
		if (video_mode >= 4)
		{
			unsigned char fg_r = palette[col & 15][0], fg_g = palette[col & 15][1], fg_b = palette[col & 15][2];
			color = SDL_MapRGB(screen->format, fg_r, fg_g, fg_b);
		}

		if (y >= height)
			break;

		for(x=off_x; x<(off_x + 8); x++)
		{
			if (x >= width)
				break;

			if (cur_pattern & 128)
			{
				if (video_mode == 8)
					setPixel(x, y, i + (col & 4 ? 127 : 0), i + (col & 2 ? 127 : 0), i + (col & 1 ? 127 : 0));
				else if (video_mode >= 4)
					setPixelUint(x, y, color);
				else
					setPixelUint(x, y, color);
			}

			cur_pattern <<= 1;
		}

		ynr++;
	}
}

void redraw_sprites_screen()
{
	int attr, patt;
	int nr;

	if (video_mode >= 4)
		attr = ((vdp_registers[5] & ~3) << 7) + ((vdp_registers[11] & 3) << 15);
	else
		attr = vdp_registers[5] << 7;

	patt = vdp_registers[6] << 11;

	if (vdp_registers[1] & 2)		/* 16 x 16 */
	{
		fprintf(stderr, "16x16 sprites\n");

		for(nr=0; nr<32; nr++)
		{
			int attribute_offset = attr + nr * 4;
			int spy = read_vdp(vdp_mem, attribute_offset + 0);
			int spx = read_vdp(vdp_mem, attribute_offset + 1);

			if (spy == 0xd0)
				break;

			if (spx < 256 && spy < 192)
			{
				int color = read_vdp(vdp_mem, attribute_offset + 3) & 15;

				if (color || video_mode >= 4)	/* color 0: transparent */
				{
					int pattern_index = read_vdp(vdp_mem, attribute_offset + 2);
					int offset = patt + (8 * pattern_index);
					unsigned char fg_r = rgb[color][0], fg_g = rgb[color][1], fg_b = rgb[color][2];
					Uint32 colorfg = SDL_MapRGB(screen->format, fg_r, fg_g, fg_b);

					draw_sprite_part(spx + 0, spy + 0, offset +  0, colorfg, nr);

					draw_sprite_part(spx + 0, spy + 8, offset +  8, colorfg, nr);

					draw_sprite_part(spx + 8, spy + 0, offset + 16, colorfg, nr);

					draw_sprite_part(spx + 8, spy + 8, offset + 24, colorfg, nr);
				}
			}
		}
	}
	else					/*  8 x  8 */
	{
		fprintf(stderr, "8x8 sprites\n");

		for(nr=0; nr<32; nr++)
		{
			int attribute_offset = attr + nr * 4;
			int spy = read_vdp(vdp_mem, attribute_offset + 0);
			int spx = read_vdp(vdp_mem, attribute_offset + 1);

			if (spy == 0xd0)
				break;

			if (spx < 256 && spy < 192)
			{
				int color = read_vdp(vdp_mem, attribute_offset + 3) & 15;

				if (color || video_mode >= 4)
				{
					int pattern_index = read_vdp(vdp_mem, attribute_offset + 2);
					unsigned char fg_r = rgb[color][0], fg_g = rgb[color][1], fg_b = rgb[color][2];
					Uint32 colorfg = SDL_MapRGB(screen->format, fg_r, fg_g, fg_b);

					draw_sprite_part(spx, spy, pattern_index, colorfg, nr);
				}
			}
		}
	}
}

void redraw_sprites_msx2()
{
	int nr;

	for(nr=0; nr<32; nr++)
	{
		int attribute_offset = sat + nr * 4;
		int spy = read_vdp(vdp_mem, attribute_offset + 0);
		int spx = read_vdp(vdp_mem, attribute_offset + 1);

		if (spy == 0xd0)
			break;
	}
}

void redraw_screen2()
{
	int bg_map    = (vdp_registers[2] &  15) << 10;
	int bg_colors = (vdp_registers[3] & 128) <<  6; /* in scr 2 only bit 7 is used, so offset is 0x2000 or 0x0 */
	int bg_tiles  = (vdp_registers[4] &   4) << 11; /* in scr 2 only bit 2 */
	int map_index, loop;
	Uint32 colors[16];

	// FIXME: kleur index 0: vdp reg 7 bg
	for(loop=0; loop<16; loop++)
		colors[loop] = SDL_MapRGB(screen->format, rgb[loop][0], rgb[loop][1], rgb[loop][2]);

	for(map_index=0; map_index<(32*24); map_index++)
	{
		int cur_char_nr = read_vdp(vdp_mem, bg_map + map_index);
		int block_nr    = (map_index >> 8) & 3;
		int cur_tiles   = bg_tiles  + (block_nr * 256 * 8) + (cur_char_nr * 8);
		int cur_colors  = bg_colors + (block_nr * 256 * 8) + (cur_char_nr * 8);
		/* offset in SDL window */
		int scr_x = (map_index & 31) * 8;
		int scr_y = ((map_index >> 5) & 31) * 8;
		int x, y;
		unsigned char *ptiles = &vdp_mem[cur_tiles];
		unsigned char *pcolors = &vdp_mem[cur_colors];
		int end_x = scr_x + 8;

		for(y=8; y; y--, scr_y++)
		{
			unsigned char current_tile  = *(ptiles++);
			unsigned char current_color = *(pcolors++);
			Uint32 fg = colors[current_color >> 4];
			Uint32 bg = colors[current_color & 15];

			for(x=scr_x; x<end_x; x++, current_tile <<= 1)
				setPixelUint(x, scr_y, current_tile & 128 ? fg : bg);
		}
	}
}

void redraw_screen5()
{
	int name_table = (vdp_registers[2] & 0x60) << 9;
	int xp, y, loop;
	Uint32 colors[16];

	for(loop=0; loop<16; loop++)
		colors[loop] = SDL_MapRGB(screen->format, palette[loop][0], palette[loop][1], palette[loop][2]);

	int p = name_table, x;
	for(y=0; y<212; y++)
	{
		for(x=0,xp=0; xp<128; xp++)
		{
			unsigned char byte = read_vdp(vdp_mem, p++);
			int xp0 = byte >> 4, xp1 = byte & 15;

			setPixelUint(x++, y, colors[xp0]);
			setPixelUint(x++, y, colors[xp1]);
		}
	}
}

void redraw_screen6()
{
	int name_table = (vdp_registers[2] & 0x60) << 9;
	int xp, y, loop;
	Uint32 colors[4];

	for(loop=0; loop<4; loop++)
		colors[loop] = SDL_MapRGB(screen->format, palette[loop][0], palette[loop][1], palette[loop][2]);

	for(y=0; y<((vdp_registers[9] & 128) ? 212 : 192); y++)
	{
		int yp = ((y + vdp_registers[0x17]) % 212) * 128;

		for(xp=0; xp<128; xp++)
		{
			int x = xp * 4;
			int p = name_table + yp + xp;
			unsigned char byte = read_vdp(vdp_mem, p);
			int xp0 = byte >> 6, xp1 = (byte >> 4) & 3, xp2 = (byte >> 2) & 3, xp3 = byte & 3;

			setPixelUint(x + 0, y, colors[xp0]);
			setPixelUint(x + 1, y, colors[xp1]);
			setPixelUint(x + 2, y, colors[xp2]);
			setPixelUint(x + 3, y, colors[xp3]);
		}
	}
}

void redraw_screen7()
{
	int name_table = (vdp_registers[2] & 0x60) << 9;
	int xp, y, loop;
	Uint32 colors[15];

	for(loop=0; loop<15; loop++)
		colors[loop] = SDL_MapRGB(screen->format, palette[loop][0], palette[loop][1], palette[loop][2]);

	for(y=0; y<((vdp_registers[9] & 128) ? 212 : 192); y++)
	{
		int yp = ((y + vdp_registers[0x17]) % 212) * 256;

		for(xp=0; xp<256; xp++)
		{
			int x = xp * 2;
			int p = name_table + yp + xp;
			unsigned char byte = read_vdp(vdp_mem, p);
			int xp0 = byte >> 4, xp1 = byte & 15;

			setPixelUint(x + 0, y, colors[xp0]);
			setPixelUint(x + 1, y, colors[xp1]);
		}
	}
}

void redraw_screen8()
{
	int name_table = 0;
	int x, y;

	for(y=0; y<212; y++)
	{
		int offset = name_table + y *256;

		for(x=0; x<256; x++)
		{
			unsigned char byte = read_vdp(vdp_mem, offset + x);

			setPixel(x, y, rg[(byte >> 2) & 7], rg[(byte >> 5) & 7], b[byte & 3]);
		}
	}
}

void redraw_graphics(unsigned char *vdp_mem, unsigned char *vdp_registers)
{
	static double last_ts = 0.0;
	static int cur_fps = 0;
	static long int frame_nr = 0;
	double now = get_ts();

	cur_fps++;

	if ((now - last_ts) >= 1.0)
	{
		last_ts = now;
		fps = cur_fps;
		cur_fps = 0;
	}

	frame_nr++;

	if (skip_frame != 0 && (frame_nr % skip_frame) == 0)
	{
		/* skip frame */
	}
	else
	{
		lockScreen();

		switch(video_mode)
		{
			case 1:			/* 40x24 */
			case 80:		/* 80x24 */
				redraw_screen0(video_mode == 1 ? 40 : 80);
				break;

			case 0:			/* 'screen 1' */
				redraw_screen1();
				break;

			case 2:			/* 256x192 */
				redraw_screen2();
				redraw_sprites_screen();
				break;

			case 5:			/* 256x212 16 colors */
				redraw_screen5();
				redraw_sprites_screen();
				break;

			case 6:			/* 512x212 4 colors */
				redraw_screen6();
				redraw_sprites_screen();
				break;

			case 7:			/* 512x212 16 colors */
				redraw_screen7();
				redraw_sprites_screen();
				break;

			case 8:			/* 256x212, 256 colors */
				redraw_screen8();
				redraw_sprites_screen();
				break;
		}

		unlockScreen();
		updateScreen();
	}
}

void resize_window(int xres, int yres)
{
	if ((screen = SDL_SetVideoMode(xres, yres, 32, (svi -> hw_available?SDL_HWSURFACE:SDL_SWSURFACE) | SDL_ASYNCBLIT)) == NULL)
		error_exit("Couldn't set %dx212 video mode: %s", xres, SDL_GetError());
}

void init_video(char *map_log)
{
	int xres = 256, yres = 192, loop;

	initscr();
	immedok(stdscr, FALSE);

	if (SDL_Init(SDL_INIT_VIDEO) < 0)
		error_exit("Couldn't initialize SDL: %s", SDL_GetError());

	atexit(SDL_Quit);

	svi = SDL_GetVideoInfo();

	if (verbose) fprintf(stderr, "bits per pixel: %d\n", svi -> vfmt -> BitsPerPixel);
	if (verbose) fprintf(stderr, "surface: %s\n", svi -> hw_available ? "hardware" : "software");
	fprintf(stderr, "BPP: %d\n", svi -> vfmt -> BitsPerPixel);
	if ((screen = SDL_SetVideoMode(xres, yres, 32, (svi -> hw_available?SDL_HWSURFACE:SDL_SWSURFACE) | SDL_ASYNCBLIT)) == NULL)
		error_exit("Couldn't set %dx212 video mode: %s", xres, SDL_GetError());

	SDL_EnableUNICODE(1);

	if (map_log)
	{
		map_log_fh = fopen(map_log, "a");
		if (!map_log_fh)
			error_exit("failed accessing file %s\n", map_log);
	}

	memset(vdp_status_regs, 0x00, sizeof(vdp_status_regs));

	for(loop=0; loop<16; loop++)
	{
		palette[loop][0] = rgb[loop][0];
		palette[loop][1] = rgb[loop][1];
		palette[loop][2] = rgb[loop][2];
	}

	for(loop=0; loop<8; loop++)
		rg[loop] = (int)((double)loop * (255.0 / 7.0));
	for(loop=0; loop<4; loop++)
		b[loop]  = (int)((double)loop * (255.0 / 3.0));
}
